package jburg.burg;

//  BURM specification is parsed into ANTLR2 AST objects.
import antlr.collections.AST;
//  The code emitter interface definition.
import jburg.emitter.EmitLang;
// The code emitter works by string templates.
import org.antlr.stringtemplate.*;

//  The factory that selects and creates an emitter.
import jburg.emitter.JBurgEmitterFactory;

//  The i-node adapter interface definition.
import jburg.burg.inode.InodeAdapter;
import jburg.burg.inode.InodeAdapter2;
//  Interface for i-node adapters that need
//  to emit support routines.
import jburg.burg.inode.InodeAuxiliarySupport;

//  The factory that selects and creates an I-node adapter.
import jburg.burg.inode.InodeAdapterFactory;

//  Token types from ANTLR.
import jburg.parser.JBurgTokenTypes;

//  Version file, generated by build.
import jburg.version.JBurgVersion;

import java.io.PrintStream;
import java.io.PrintWriter;

import java.lang.reflect.Modifier;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;


@SuppressWarnings("nls")
public class JBurgGenerator implements JBurgTokenTypes
{
    /**
     *  The JBurg specification's rules are categorized as:
     *  <ul>
     *  <li> Pattern  (e.g., integer = PLUS(integer, integer) )
     *  <li> Simple Transformational   (e.g., registerOperand  = integer)
     *  <li> Complex Transformational  (e.g., string = integer { code to convert integer to string} )
     *  </ul>
     *
     *  Simple transformational rules allow a subgoal to satisfy
     *  other subgoals without additional processing; complex
     *  transformational rules allow one subgoal to feed additional
     *  processing that can satisfy other subgoals (at added cost).
     *
     *  Syntactically, a simple transformational rule is distinguished 
	 *  by its lack of a block of code.
	 *  Simple transformational rules are also known as "chain rules"
	 *  in simpler BURGs that don't support complex transformational
	 *  rules.  This longer name was adopted to highlight that difference,
	 *  but a simple transformational rule is, in fact, a chain rule
	 *  by a different name.
     */

	/**
     *   The table of all transformation rules; keyed
     *   by the rules' necessary state, e.g., integer
	 *   in the rule number = integer).
     */
    Map<String, Vector<JBurgRule>> transformationRules = new HashMap<String, Vector<JBurgRule>>();

	/**
	 *  The table of all pattern rules, keyed
	 *  by the top-level operator of the pattern,
	 *  e.g., PLUS in the pattern integer = PLUS(integer, integer).
	 */
    Map<String, Vector<JBurgRule>> patternRules = new HashMap<String, Vector<JBurgRule>>();

    /**
     *  The goal states' names become symbolic constants
     *  in the generated reducer or the nonterminal enumeration.
     */
    Set<String> allGoalStates = new HashSet<String>();

    /**
     *  Closure sets are organized by their goal state; each state's
     *  closure set contains an entry for every closure
     *  path from any other non-terminal state to the target state.
     */
    Map<String, Vector<ClosureRecord>> closureSets = new HashMap<String, Vector<ClosureRecord>>();

    /**  Action code fragments.  */
    ArrayList<JBurgReduceAction> reduceActions = new ArrayList<JBurgReduceAction>();

    /**
     *  Cost function fragments.
     *  Note that cost functions may also be embedded
     *  in the inline code blocks in a specification.
     */
    ArrayList<AST> costFunctions = new ArrayList<AST>();

    /**  The name of the generated BURM's base class. */
    String baseClassName = null;

    /**  The names of any interfaces that the generated BURM implements.  */
    Vector<String> interfaceNames = new Vector<String>();
    
    /** blocks of code to be added into the class verbatim */
    Vector<String> inclassCode = new Vector<String>();

    /**
     *  Rules that can generate compressed annotations.
     */
    RulesByOperatorAndArity compressedAnnotations = new RulesByOperatorAndArity();

    /** The package name of the generated reducer. */
    String packageName = null;

    /**  Header code, copied as-is into the reducer. */
    String headerBlock = null;

    /**
     *  The name of the i-node class that's being labeled and reduced.
     */
    String iNodeClass = null;

	/** 
	 *  I-node adapter class name.  The adapter is instantiated by name.
	 *  @note: specified as an alternative to the I-node class' name,
	 *  which selects a builtin adapter.
	 */
	String iNodeAdapterClass;

    /**
     *  The return types for specific states.
     */
    Map<String, String> returnTypeTable = new HashMap<String, String>();

    /**
     *  The default return type of the reducer functions.
     *  If this is defaulted to null, the generated reducer will return nodes of the iNodeClass.
     */
    String defaultReturnType = null;

    /**
     *  Table of properties to be added to the BURM.
     *  Each one is a name/type pair; the BURM gets
     *  a private member and get/set methods.
     */
    Map<String, String> burmProperties = new HashMap<String, String>();

    /**
     *  Simple transformations simply delegate to their antecedent reduction,
     *  so all transformations to a given nonterminal state can share the
     *  same rule.
     */
    Map<String, JBurgRule> simpleTransformationRules = new HashMap<String,JBurgRule>();

    /**  Include debugging code in the generated reducer?  */
    boolean debugMode;
    
    /** 
     * If the pattern-matcher generator fails, dump its annotation tree here.
     * Note: this is only enabled (or useful) when debugging JBurg's own BURM. 
     */
    String patternMatcherDumpFile = null;

    /**  
     *  Caller's logging interface.  If defaulted to null, informational and
     *  error messages go to System.out and System.err respectively.
     */
    Logger logger = null;
    
    /**
     * Name for the language to emit code in (default is assumed to be Java)
     */
    String emitLanguageName = null;
    
    /** Code emitter to use (selected using the language name) */
	EmitLang codeEmitter = null;

	/** I-node adapter to use */
	InodeAdapter iNodeAdapter;
	
	/** 
	 *  If the InodeAdapter is also an InodeAdapter2,
	 *  keep a reference to it. 
	 */
	InodeAdapter2 adapter2;

	/**  Cumulative error count. */
	int errCount = 0;

	/**  Name of the initial parameter to the label routine. */
	private static String initialParamName = "to_be_labelled";
	
	/** Name of the node in the reducer */
	private String reducerNodeName = "__p";
	
	/** Name of the stack of reduced values */
	private String reducedValuesName = "__reducedValues";

    /**  Default error handler.  null means use hard-coded default, i.e., throw an exception. */
    String defaultErrorHandler = null;

    /**  Prologue snippets mapped to the corresponding rule number. */
    Map<Integer,AST> prologueBlocks = new TreeMap<Integer,AST>();

    /** 
     *  All operators known to any pattern accumulate in this set
     *  so that the BURG can generate operator-specific getNthChild()
     *  routines.
     */
    Set<String> allOperators = new TreeSet<String>();
    
    /**
     * operator mapped to node type. This map is populated by JBurg.NodeType
     * directives.
     * <p>
     * For example, an operator named IdentifierID would be associated with
     * {@code IIdentifierNode*} in this map if the input contained
     * {@code JBurg.NodeType IdentifierID = IIdentifierNode*;
     */
    final Map<String, String> opcodeNodeTypes = new HashMap<String, String>();
    
    /**
     *  The type of the INodes' opcode.  Defaults to int but
     *  can be an enum for maintainability.
     */
    String opcodeType = "int";
    
    /**
     * The type of the generated nonterminals.  Defaults to int
     * but can be an enum for maintainability and error detection.
     */
    String ntType = "int";

    /**
     *  Manifest constants in the JBurg specification.
     */
    Map<String,Integer> manifestConstants = new HashMap<String,Integer>();
    
    /**
     *  Patterns that have been optimized out are represented
     *  by this not-likely-to-compile sequence.
     */
    final private static String nilPattern = "-null-";

    /**
     *  Manifest constant for method declarations.
     */
    final private static Class<?>[] throwsNothing = null;

    /**
     *  Manifest constant for method declarations.
     */
    final private static Class<?>[] throwsException = new Class<?>[] { Exception.class };

    /**
     *  Manifest constant for method declarations.
     */
    final private static String[][] noFormalParameters = new String[][] { };

    /**
     *  Manifest constant for an unfeasible rule.
     */
    final private static String NO_FEASIBLE_RULE = "-1";

    /**
     *  Manifest constant for an uninitialized cost.
     */
    final private static String UNINITIALIZED = "-1";

    /**
     *  NamedPattern holds a named pattern, and keeps a table of the reductions that refer to it.
     *  (See defect 3063143 for problems with multiple reductions sharing a pattern.)
     */
    class NamedPattern
    {
        String patternName;
        AST pattern = null;
        Vector<AST> reductions = new Vector<AST>();

        NamedPattern(String name)
        {
            this.patternName = name;
        }
    }

    /**
     *  A PatternMap keeps the map of pattern names to reductions.
     */
    @SuppressWarnings("serial")
    class PatternMap extends  TreeMap<String,NamedPattern>
    {
        NamedPattern getPattern(String key)
        {
            if ( !super.containsKey(key) )
                put(key, new NamedPattern(key));
            return super.get(key);
        }

        void addPattern(String key, AST pattern)
        {
            NamedPattern p = getPattern(key);
            p.pattern = pattern;
        }

        void addReduction(String key, AST reduction)
        {
            NamedPattern p = getPattern(key);
            p.reductions.add(reduction);
        }
    }

    /**
     * Patterns declared separately from the
     * reduction(s) that use them, e.g.,
     * Pattern callMethod = CALL(expression method, expression parameters*);
     * expression = Pattern callMethod: cost1
     * void_expression = Pattern callMethod: cost2
     */
    PatternMap namedPatterns = new  PatternMap();

    /**
     *  Aggregate path computations if more than this threshold
     *  number of pattern elements share a path.
     */
    private int aggregatePathThreshold = Integer.MAX_VALUE;

    public void setAggregatePathThreshold(int threshold)
    {
        this.aggregatePathThreshold = threshold;
    }

    /**
     * Cache closure computations if the number of
     * elements in the computation exceeds this threshold.
     */
    private int closureCacheThreshold = 7;

    /**
     * Volatile cost functions.
     */
    List<String> volatileCostFunctions = new ArrayList<String>();

    /**
     * Node types that require agressive caching.
     */
    List<String> aggresiveCacheTypes = new ArrayList<String>();

    /**
     *  @param root - the root of the AST generated by parsing the .jburg specification.
     *  @param log - a sink for diagnostics.
     */
    public JBurgGenerator(AST root, Logger log) throws Exception
    {
        this.logger = log;

        //  Walk over the children of the root, each of which 
		//  is a self-contained syntactical construct, and
        //  perform the appropriate action for each.
        for (AST currentNode = root; currentNode != null;
                currentNode = currentNode.getNextSibling()) 
		{
            switch (currentNode.getType()) 
			{
            case COST_FUNCTION:
                this.costFunctions.add(currentNode);
                break;

            case HEADER_DECLARATION:

				if ( null == this.headerBlock )
				{
					this.headerBlock = getCodeBlock(currentNode);
				}
				else
				{
                    throw new IllegalArgumentException("The class header may only be specified once.");
				}

                break;
                
            case INCLASS_DECLARATION:
				{
            		this.inclassCode.add( stripBrackets(getCodeBlock(currentNode)) );
            	}
				
				break;

            case INODE_ADAPTER_DECLARATION:

                if (this.iNodeAdapterClass == null)
				{
					this.iNodeAdapterClass = getIdentifierText(currentNode.getFirstChild());
				}
				else
				{
                    throw new IllegalArgumentException("INodeAdapter may only be specified once.");
				}

				break;

            case INODE_TYPE_DECLARATION:

                if (this.iNodeClass == null)
				{
                    this.iNodeClass = getIdentifierText(currentNode.getFirstChild());
				}
                else
				{
                    throw new IllegalArgumentException("INodeType may only be specified once.");
				}

                break;
                
            case LANGUAGE_DECLARATION:

            	if ( null == this.emitLanguageName )
				{
            		this.emitLanguageName = currentNode.getFirstChild().getText();
				}
            	else
				{
            		throw new IllegalArgumentException("Language may only be specified once.");
				}

            	break;

            case EXTENDS_CLASS_SPECIFICATION:
                this.baseClassName = getIdentifierText( currentNode.getFirstChild());
                break;

            case IMPLEMENTS_INTERFACE_SPECIFICATION:
                this.interfaceNames.addElement(getIdentifierText( currentNode.getFirstChild()) );

                break;

            case PACKAGE_SPECIFICATION:
                if ( null == this.packageName )
				{
					this.packageName = getIdentifierText(currentNode.getFirstChild());
				}
				else
				{
            		throw new IllegalArgumentException("package may only be specified once.");
				}

                break;

            case PROPERTY_SPECIFICATION: 
				{
					String propertyType = getIdentifierText(currentNode.getFirstChild());
					String propertyName = currentNode.getFirstChild().getNextSibling().getText();
					this.burmProperties.put(propertyName, propertyType);
				}

                break;

            case RETURN_DECLARATION:

                if (this.defaultReturnType == null)
                    this.defaultReturnType = getIdentifierText(currentNode.getFirstChild());
                else
                    throw new IllegalArgumentException( "ReturnType may only be specified once.");

                break;

            case PATTERN_RULE:
                addPatternRule(currentNode);

                break;

            case SIMPLE_TRANSFORMATION_RULE:
                addSimpleTransformationRule(currentNode);

                break;

            case TRANSFORMATION_RULE:
                addComplexTransformationRule(currentNode);

                break;

            case TYPED_RETURN_DECLARATION:

                {
                    String stateName = currentNode.getFirstChild().getText();
                    String returnType = getIdentifierText(currentNode.getFirstChild()
                            .getNextSibling());

                    //  Put the return declaration in the table, but only once per state.
                    Object typeCollision = this.returnTypeTable.put(stateName, returnType);
                    if ( null != typeCollision )
                    {
                        throw new IllegalArgumentException(
                            "A state may only specify one ReturnType."
                        );
                    }
                }

                break;

            case PATTERN_DECLARATION:
                {
                    String pattern_name = currentNode.getFirstChild().getText();
                    namedPatterns.addPattern(pattern_name, currentNode);
                }
                break;

            case REDUCTION_DECLARATION:
                {
                    String pattern_name = currentNode.getFirstChild().getNextSibling().getText();
                    namedPatterns.addReduction(pattern_name, currentNode);
                }
                break;

            case DEFAULT_ERROR_HANDLER:
                this.defaultErrorHandler = getCodeBlock(currentNode);
                break;
            case NODE_TYPE:
            {
                final String operatorID = currentNode.getFirstChild().getText();
                assert !operatorID.isEmpty() : "Parser should never create empty operator!";
                final String nodeType = currentNode.getFirstChild().getNextSibling().getText();               
                assert !nodeType.isEmpty() : "Parser should never create empty node type!";
                
                if (this.opcodeNodeTypes.put(operatorID, nodeType) != null)
                {
                    final String message = "Duplicate node type declaration for '"
                        + operatorID
                        + "'.";
                    throw new IllegalArgumentException(message);
                }
                break;
            }
            case OPCODE_TYPE:
                this.opcodeType = currentNode.getFirstChild().getText();
                break;

            case NON_TERMINAL_ENUM:
                this.ntType = currentNode.getFirstChild().getText();
                break;

            case MANIFEST_CONSTANT:
                manifestConstants.put(currentNode.getFirstChild().getText(), Integer.parseInt(currentNode.getFirstChild().getNextSibling().getText()));
                break;

            case VOLATILE_COST:
            {
                for ( AST volatileCost = currentNode.getFirstChild().getNextSibling(); volatileCost != null; volatileCost = volatileCost.getNextSibling() )
                {
                    volatileCostFunctions.add(volatileCost.getText());
                }
                break;
            }

            default:
                throw new IllegalArgumentException("Unknown specification AST type " +
            											String.valueOf(currentNode.getType()));
            }
        }

        //  Set the language emitter.
        codeEmitter = JBurgEmitterFactory.getEmitter(emitLanguageName, getLogger());

        if ( codeEmitter == null )
		{
			throw new IllegalStateException("Unknown language specified: \""+ emitLanguageName +"\"");
        }

        // If an i-node adapter was specified, instantiate an instance.
		if ( iNodeClass != null )
		{
			if ( iNodeAdapterClass != null )
			{
				try
				{
					this.iNodeAdapter = (InodeAdapter)Class.forName(iNodeAdapterClass).newInstance();
				}
				catch ( Exception ex )
				{
					this.logger.exception("loading i-node adapter", ex);
                    this.logger.error("Unable to instantiate i-node adapter %s", iNodeAdapterClass);
				}
			}
			else
			{
				this.iNodeAdapter = InodeAdapterFactory.getAdapter(iNodeClass);
			}
			
			codeEmitter.setINodeType(this.iNodeClass);

			if ( this.iNodeAdapter != null )
			{
				logger.info("Using i-node adapter %s", this.iNodeAdapter.getClass().getName() );
			}
			else
			{
				getLogger().warning("Using default i-node adapter, no adapter matches %s", iNodeClass );
				this.iNodeAdapter = new jburg.burg.inode.DefaultAdapter();
			}
			
			//  See if the adapter is also an InodeAdapter2 implementation.
			this.adapter2 = (this.iNodeAdapter instanceof InodeAdapter2)? (InodeAdapter2)this.iNodeAdapter : null;
		}
		else
		{
			logger.error("You must specify the i-node type.");
		}

        if ( logger.getErrorCount() > 0 )
            return;

		codeEmitter.setInodeAdapter(this.iNodeAdapter);
        codeEmitter.setNtType(this.ntType);

        //  Default return type is the same as the INode class.
        if (this.defaultReturnType == null)
		{
            this.defaultReturnType = this.iNodeClass;
		}

        //  Mutate pattern/reduction decl pairs into pattern rules.
        for ( NamedPattern np: namedPatterns.values() )
        {
            if ( np.pattern != null && np.reductions.size() > 0 )
            {
                AST named_pattern = np.pattern.getFirstChild().getNextSibling().getFirstChild();

                for ( AST reduction: np.reductions )
                {    
                    //  Splice the pattern into the reduction AST.
                
                    AST nt_state = reduction.getFirstChild();
                    AST pattern_name = nt_state.getNextSibling();
                    AST cost_decl = pattern_name.getNextSibling();
                    
                    //  Create a new holder for the pattern
                    //  so the original AST isn't mutated.
                    AST pattern_holder = new antlr.CommonAST();
                    pattern_holder.setFirstChild(named_pattern);
                    nt_state.setNextSibling(pattern_holder);
                    pattern_holder.setNextSibling(cost_decl);
    
                    //  Give the composite AST the appropriate type
                    //  for its pattern.
                    reduction.setType(PATTERN_RULE);
    
                    addPatternRule(reduction);
                }
            }
            else if ( np.pattern != null )
            {
                getLogger().warning("pattern %s has no reduction - ignored.", np.patternName );
            }
            else if ( np.reductions.size() > 0 )
            {
                throw new IllegalStateException("Reduction " + np.patternName + " has no associated pattern.");
            }
        }

        //  Add target-specific logic to the simple transformations' rules.
        for ( JBurgRule rule: this.simpleTransformationRules.values() )
        {
            JBurgReduceAction action = addAction(rule, rule.getGoalState());
            action.setAntecedentState(rule.getAntecedentState());
        }

        //  Compute the closure sets.
        computeClosureSets();
    }
    
    /**
     *  Add a reduce action to the action list.
	 *  Actions are keyed by number in the generated BURM.
     *  @param action - the rule that uses this action.
     *  @param goalState - the nonterminal state that represents
     *  the result of this reduction.
     */
    private JBurgReduceAction addAction(JBurgRule rule, String goalState)
    {
        JBurgReduceAction newAction = new JBurgReduceAction(rule, goalState);
        rule.setReduceAction(newAction);
        this.reduceActions.add(newAction);

        //  Annotate this action with its state index:
        //  the 0th index is used as the default action for
        //  simple transformation rules, so use 1-based counting.
        newAction.setIndex(this.reduceActions.size());
        return newAction;
    }

    /**
     *  Analyze a rule's action routine specifications
     *  and extract its prologue reduction, if any.
     *  @param rule - rule.
     *  
     */
    private void extractPrologue(JBurgRule rule)
    {
        AST reduction = null;

        if ( hasASTOfType(rule.m_AST, REDUCTION_ACTION) )
            reduction = getASTByType(rule.m_AST, REDUCTION_ACTION);
        else if ( hasASTOfType(rule.m_AST, EXPLICIT_REDUCTION ) )
            reduction = getASTByType(rule.m_AST, EXPLICIT_REDUCTION);
        
        if ( hasASTOfType(reduction, PROLOGUE) )
            this.prologueBlocks.put(rule.getReduceAction().index, getASTByType(reduction, PROLOGUE));
    }

    /**
     *  Add a complex transformation rule to its table,
     *  and record the rule's reduction action.
     */
    private void addComplexTransformationRule(AST transform)
        throws Exception
    {
        JBurgRule n = addNamedRule(this.transformationRules, transform);

        //  Prepare the rule's reduce action.
        JBurgReduceAction action = addAction(n, n.getGoalState());
        action.setAntecedentState(n.getAntecedentState());
        extractPrologue(n);

        //  Add the antecedent state by name as an alias action routine "parameter."
        //  The "parameter" is popped off the reduced values stack into a local variable.
        action.addParameter(
            n.getAntecedentState(), 
            n.getAntecedentState(), 
            ParameterDescriptor.ArityType.Fixed
        );
    }

    /**
     *  If this simple transformation is not yet known, add it to its rule table.
     */
    private void addSimpleTransformationRule(AST transform)
    {
        JBurgRule newRule = new JBurgRule(transform);

        String key = String.format("%s=%s", newRule.getGoalState(), newRule.getAntecedentState());

        if ( ! this.simpleTransformationRules.containsKey(key) )
        {
            //  Ensure this rule's nonterminal is known.
            this.allGoalStates.add(newRule.getGoalState());
            addNamedRule(this.transformationRules, newRule);

            //  Target-specific action logic will be
            //  added once the code emitter is up.
            this.simpleTransformationRules.put(key, newRule);
        }
    }

    /**
     *  Wrap a pattern rule in an I-node, and add it to the pattern rule table.
     */
    private void addPatternRule(AST newRule) throws Exception
    {
        JBurgRule n = addNamedRule(this.patternRules, newRule);
        addAction(n, n.getGoalState());
        extractPrologue(n);
    }

    /**
     *  Wrap a rule in an I-node, and add it to the list of rules
     *  that produce a particular subgoal.
     */
    private JBurgRule addNamedRule(Map<String, Vector<JBurgRule>> ruleTable, AST newAST)
    {
        JBurgRule newRule = new JBurgRule(newAST);

        //  Ensure this rule's nonterminal is known.
        this.allGoalStates.add( newRule.getGoalState() );
        return addNamedRule(ruleTable, newRule);
    }

    /**
     *  Add a rule to its rule table.
     *  @param ruleTable - the appropriate table for this type of rule.  Keyed by operator.
     *  @param newRule - the rule to add.
     */
    private JBurgRule addNamedRule(Map<String, Vector<JBurgRule>> ruleTable, JBurgRule newRule)
    {
        //  Store the rule by operator.
        String operator = newRule.getOperator();

        Vector<JBurgRule> vRules = ruleTable.get(operator);

        if (vRules == null)
        {
            vRules = new Vector<JBurgRule>();
            ruleTable.put(operator, vRules);
        }

        vRules.add(newRule);

        return newRule;
    }

    /**
     *  Compute the closure set of a rule:  the set of rules that
     *  that can transform this rule's goal to satsify another goal.
	 *  @note computeClosure is public so that it can be applied
	 *    by JBurgUtilities.
     */
    public void computeClosure(JBurgRule n) throws Exception
    {
		//  Get the list of transformation rules that can use this rule's goal state.
        Vector<JBurgRule> vClosureCandidate = this.transformationRules.get(n.getGoalState());

        if (vClosureCandidate != null) 
		{
            //  Found some closures:
            //  create or fetch this target state's closure set.
            Vector<ClosureRecord> vClosureSet = this.closureSets.get(n.getGoalState());

            if (vClosureSet == null) 
			{
                vClosureSet = new Vector<ClosureRecord>();
                this.closureSets.put(n.getGoalState(), vClosureSet);
            }

            for (JBurgRule nPrime:vClosureCandidate ) 
			{

                //  Add this closure to the closure set.
                ClosureRecord newClosure = new ClosureRecord(nPrime);

                if (!vClosureSet.contains(newClosure)) 
				{
                    vClosureSet.add(newClosure);
                }
            }
        }
    }

    /**
     *  Scan all the rules and compute their closure sets.
     */
    private void computeClosureSets() throws Exception
    {
        for (Vector<JBurgRule> patterns : this.patternRules.values() ) 
		{
            try 
			{
                JBurgUtilities.applyToAll( this, patterns, "computeClosure", JBurgRule.class);
            }
			catch (Exception e)
			{
                e.printStackTrace();
            }
        }

        for (Vector<JBurgRule> transformations : this.transformationRules.values() )
		{
            try
			{
                JBurgUtilities.applyToAll( this, transformations, "computeClosure", JBurgRule.class);
            }
			catch (Exception e)
			{
                e.printStackTrace();
            }
        }
    }

    /**
     *  Emit the BURG specification's resultant BURM to an output stream.
     */
    public int generate(String className, PrintStream output)
    {
        try
		{
            codeEmitter.setOpcodeType(this.opcodeType);
            codeEmitter.setDefaultAttribute("className", null);
        	emitHeader(className, output);
        	codeEmitter.emitInclass(iNodeClass, this.inclassCode, output);

        	emitNTConstants(this.allGoalStates, output);

			if ( this.iNodeAdapter instanceof InodeAuxiliarySupport )
			{
				((InodeAuxiliarySupport)this.iNodeAdapter).emitAuxiliarySupport(codeEmitter, output);
			}

            emitLabelFunction(output);
            ruleSemanticAnalysis();

            if  ( codeEmitter.supportsSpecializedAnnotations() )
            {
                for (Vector<JBurgRule> vPatternRules: this.patternRules.values())
                    this.compressedAnnotations.addAll(vPatternRules);
            }
            else
            {
                emitComputeCostMatrixFunction(this.iNodeClass, output);
                emitClosures(this.closureSets, this.iNodeClass, output);
            }
            emitActions(this.reduceActions, this.iNodeClass, output);
            emitCostFunctions(this.costFunctions, this.iNodeClass, output);

            // Eventually this will be the only way to emit annotations,
            // but until all targets are ported it must be guarded.
            if ( codeEmitter.supportsSpecializedAnnotations() )
            {
                emitCompressedAnnotations(output);
                // Reset the default class name.
                codeEmitter.setDefaultAttribute("className", null);
            }
            
            //  If there's an InodeAdapter2 present,
            //  emit the getNthChild routine that
            //  implements the adapter's node-handling logic.
            if ( this.adapter2 != null)
                emitGetNthChild(output);

            //  Translate prologue blocks.
            Map<Integer, String> translatedPrologues = new HashMap<Integer, String>();

            for ( Integer key: this.prologueBlocks.keySet() )
            {
                AST prologue = this.prologueBlocks.get(key);

                if ( hasASTOfType(prologue, BLOCK) )
                    translatedPrologues.put(key, getCodeBlock(prologue));
                else if ( hasASTOfType(prologue, PROCEDURE_CALL) )
                    translatedPrologues.put(key, emitExpression(getASTByType(prologue, PROCEDURE_CALL)) + ";");
                else
                    throw new IllegalStateException("Prologue block with no inline code or callback");
            }

            emitStatics(output);
                
            codeEmitter.emitTrailer(
				className,
				this.iNodeClass,
				this.allGoalStates,
				this.burmProperties,
				this.debugMode,
                this.defaultErrorHandler,
                translatedPrologues,
				output
			);
        }
		catch (Exception e)
		{
            e.printStackTrace();
			this.errCount++;
        }

		return this.errCount;
    }

    /**
     * Emit static lookup tables.
     */
    private void emitStatics(PrintStream output)
    {
        //  Emit the static table of subgoal records.
        Map<Integer, Vector<JBurgPatternMatcher>> rules_by_action = new HashMap<Integer,Vector<JBurgPatternMatcher>>();
        int max_action = 0;
        for (Vector<JBurgRule> vPatternRules: patternRules.values())
        {
            for (JBurgRule p: vPatternRules)
            {
                int action_number = p.getReduceAction().getIndex();
                max_action = Math.max(max_action, action_number);
                
                if ( p.patternMatcher.isNary() )
                {
                    rules_by_action.put(action_number,new Vector<JBurgPatternMatcher>());
                    rules_by_action.get(action_number).add(p.patternMatcher);
                }
                else if (!p.patternMatcher.getParameterizedSubtrees().isEmpty() )

                {
                    rules_by_action.put(action_number,p.patternMatcher.getParameterizedSubtrees());
                }
            }
        }

        //  Ensure the transformation rules' actions have subgoal records.
        for ( Vector<JBurgRule> transformations:  this.transformationRules.values() )
            for ( JBurgRule t: transformations )
                max_action = Math.max(max_action, t.getReduceAction().getIndex());
        
        codeEmitter.emitStatics(max_action, rules_by_action, output);

        //  Generate manifest constant declarations.
        for ( Map.Entry<String,Integer> constants : this.manifestConstants.entrySet() )
            genInstanceField(
                output,
                Modifier.PUBLIC + Modifier.STATIC + Modifier.FINAL,
                "int",
                constants.getKey(),
                constants.getValue().toString()
            );
    }

    /**
     *  Get the payload of an identifier, with error checking.
     *  @return the identifier's text.
     */
    private String getIdentifierText(AST p)
    {
		if ( p.getType() != IDENTIFIER )
		{
			throw new IllegalStateException ( "Expected IDENTIFIER, found " + p.toStringTree() );
		}

        return p.getText();
    }

    /**
     *  Does this production have a closure set?
     *  @return true if n has a closure set.
     *  @param n - the production of interest.
     */
    public boolean hasClosure(JBurgProduction n)
    {
        return this.closureSets.get(n.getGoalState()) != null;
    }

    /**
      *  Emit the file/class header information.
      */
    private void emitHeader(String className, PrintStream output)
    {
        genComment(output, " Generated " + new java.util.Date().toString() + " by JBurg version " + JBurgVersion.version  );
        output.println();
        
        codeEmitter.emitHeader(
            className,
            this.packageName,
            this.headerBlock,
            this.baseClassName,
            this.interfaceNames,
            this.debugMode,
            output
        );
    }

	/**
	 *  @return the return type for a specific state.
	 */
	public String getReturnType(String stateName)
	{
		Object result = returnTypeTable.get(stateName);

		if (result == null)
			result = defaultReturnType;

		return result.toString();
	}
	
    /**
     *  setDebugMode controls the level of debugging code that will be generated in the reducer.
     */
    public void setDebugMode(boolean debugMode)
    {
        this.debugMode = debugMode;
    }
    
	/**
	 *  @return the text of the parent AST's code block, which is represented as a BLOCK token.
	 */
	private String getCodeBlock( AST parent )
	{
		return getASTByType(parent, BLOCK).getText();
	}

    /**
     *  @return the first child of the parent AST with the specified node type.
     */
    private AST getASTByType(AST parent, int node_type)
    {
        for ( AST current = parent.getFirstChild(); current != null; current = current.getNextSibling() )
        {
            if ( current.getType() == node_type )
                return current;
        }

        throw new IllegalStateException ( "AST " + parent.toString() + " has no child of node type " + node_type + "." );
    }

    /**
     *  @return true if the parent AST has a child of the specified node type.
     */
    private boolean hasASTOfType(AST parent, int node_type)
    {
        if ( parent == null )
            return false;
        for ( AST current = parent.getFirstChild(); current != null; current = current.getNextSibling() )
        {
            if ( current.getType() == node_type )
                return true;
        }

        return false;
    }

    /**
     * @param cost a call to a cost function.
     * @return true iff the cost function is not listed as volatile.
     */
    private boolean isStableCostFunction(String cost)
    {
        for ( String volatileCost: this.volatileCostFunctions )
            if ( cost.startsWith(volatileCost) )
                return false;

        return true;
    }

    /**
     *  A JBurgProduction is a view of a pattern-matching rule (JBurgRule) 
     *  or a nonterminal transformation rule (ClosureRecord) that exposes
     *  characteristics important in their static analysis.
     */
    public interface JBurgProduction
    {
        /**
         *  @return the nonterminal this production produces.
         */
        public String getGoalState();

        /**
         *  @return the code snippet that computes
         *    or retrieves this production's 
         *    (potentially) cached code.
         */
        public String getCachedCost();

        /**
         *  @return this production's reduce action.
         */
        public JBurgReduceAction getReduceAction();

        /**
         *  Is this production's cost constant in the
         *  context where it's to be invoked?
         *  @param productions - the set of productions at the
         *    point where this productions's cost is required.
         */
        public boolean computesConstantCost(Multimap<String, JBurgProduction> productions);

        /**
         *  Get the constant cost of a production.
         *  @param productions - the set of productions at the
         *    point where this production's cost is required.
         *  @return the cost, with overflow-safe addition.
         *  @pre computesConstantCost(productions) must be true.
         */
        public int getConstantCost(Multimap<String, JBurgProduction> productions);
    }

    /**
     *  JBurgRule contains an AST that represents a rule, 
     *  its associated JBurgReduceAction,
     *  and accessors to the AST's components.
     */
    public class JBurgRule implements JBurgProduction
    {
        /**
         *  The parsed rule.
         */
        final AST m_AST;
        
        /**
         *  If the rule is a pattern rule, its pattern matcher.
         */
        JBurgPatternMatcher patternMatcher = null;
        
        /**
         * The rule's reduction action.
         */
        JBurgReduceAction reduceAction;

        /**
         *   The pattern as a String; precomputed so it can be used as a key
         *   to group labeling choices by common patterns.
         */
        private String encodedPattern;

        public JBurgRule(AST n)
        {
            m_AST = n;
            
            if ( m_AST.getType() == PATTERN_RULE )
            {
                AST pattern_root = m_AST.getFirstChild().getNextSibling().getFirstChild();
                try
                {
                    this.patternMatcher = generateMatcher(pattern_root); 
                }
                catch ( Exception ex )
                {
                    logger.exception("Building pattern recognizer for " + m_AST.toStringTree(), ex);
                }
            }
        }

        String getEncodedPattern()
        {
            if ( this.encodedPattern == null )
            {
                this.encodedPattern = 
                    patternMatcher.generatePatternRecognizer(
                        codeEmitter,
                        "node",
                        JBurgGenerator.this.adapter2
                    );
            }
            
            if ( this.encodedPattern == null )
                this.encodedPattern = nilPattern;

            return this.encodedPattern;
        }

        /**
         * @param baseNode - the path to the base of the subtree.
         * @return the subtree's cost specification.
         */
        public String getCost(String baseNode)
        {
            if ( m_AST.getType() != SIMPLE_TRANSFORMATION_RULE )
            {
                AST cost_spec = m_AST.getFirstChild().getNextSibling()
                        .getNextSibling();

                String costText = cost_spec.getFirstChild().getText();

                if ( hasCostFunction() )
                {
                    if ( cost_spec.getFirstChild().getNextSibling() != null )
                        return emitExpression(getASTByType(m_AST, PROCEDURE_CALL));
                    else
                        return JBurgGenerator.this.codeEmitter.genCallMethod(null, costText, baseNode);
                }
                else
                    return costText;
            }
            else
            {
                // Simple transformation rules don't cost anything.
                return "0";
            }
        }
        
        /** 
         *  @return true if the rule has a constant cost.
         */
        public boolean hasConstantCost()
        {
            if ( m_AST.getType() != SIMPLE_TRANSFORMATION_RULE )
            {
                AST cost_spec = m_AST.getFirstChild().getNextSibling()
                        .getNextSibling();

                String costText = cost_spec.getFirstChild().getText();

                boolean ownCostConstant =
                    cost_spec.getType() == LITERAL_COST_SPEC ||
                    JBurgGenerator.this.manifestConstants.containsKey(costText);
                
                return ownCostConstant &&
                     (this.isTerminalPattern() || this.patternMatcher == null);
            }
            else
            {
                // Simple transformation rules all cost 0.
                return true;
            }
        }
        
        /**
         *  @return the rule's cost.
         *  @throws IllegalStateException if the cost isn't constant.
         */
        public Integer getConstantCost()
        {
            if ( m_AST.getType() != SIMPLE_TRANSFORMATION_RULE )
            {
                AST cost_spec = m_AST.getFirstChild().getNextSibling()
                        .getNextSibling();

                String costText = cost_spec.getFirstChild().getText();

                if ( cost_spec.getType() == LITERAL_COST_SPEC )
                {
                    return new Integer(costText);
                }
                else if ( JBurgGenerator.this.manifestConstants.containsKey(costText) )
                {
                    return JBurgGenerator.this.manifestConstants.get(costText);
                }
                else
                {
                    throw new IllegalStateException("non constant cost: " + costText);
                }
            }
            else
            {
                // Simple transformation rules don't cost anything.
                return 0;
            }
        }

        /**
         *  @return true if the cost specification is a function call.
         */
        public boolean hasCostFunction()
        {
            boolean result = false;

            if ( m_AST.getType() != SIMPLE_TRANSFORMATION_RULE )
                result = m_AST.getFirstChild().getNextSibling().getNextSibling().getType() == PROCEDURE_CALL;

            return result;
        }

        /**
         *  @return the rule's cached cost.
         */
        @Override
        public String getCachedCost()
        {
            return getCachedCost(reducerNodeName);
        }
        
        /**
         * @param baseName the name of the base of the dingus.  
         */
        public String getCachedCost(String baseName)
        {
            if ( this.hasCostFunction() && this.isStableCostFunction() )
                return String.format("cachedCost_%h()", getCost(reducerNodeName));
            else
                return getCost(baseName);

        }

        /**
         * @return true if the rule's cost is not volatile.
         */
        public boolean isStableCostFunction()
        {
            return JBurgGenerator.this.isStableCostFunction(getCost(reducerNodeName));
        }

        /**
         *  @return the rule's reduction action.
         */
        @Override
        public JBurgReduceAction getReduceAction()
        {
			if ( null == this.reduceAction )
			{
				throw new IllegalStateException( "getReduceAction() has no action to return." );
			}

            return this.reduceAction;
        }

        /**
         *  @return the antecedent state of a nonterminal-to-nonterminal
         *    reduction, i.e., the state that the subtree must be reduced
         *    to before this reduction can run.
         */
        public String getAntecedentState()
        {
             if ( m_AST.getType() == SIMPLE_TRANSFORMATION_RULE )
             {
                 return m_AST.getFirstChild().getNextSibling().getText();
             }
             else if ( m_AST.getType() == TRANSFORMATION_RULE )
             {
                 return m_AST.getFirstChild().getNextSibling().getText();
             }
             else
                 throw new IllegalStateException(String.format("no antecedent for %s", m_AST.getType()));
        }

        /**
         *  @return the node ID of the node at the root of the subtree.
         */
        public String getOperator()
        {
			int type = m_AST.getType();

			if ( PATTERN_RULE == type )
			{
				return m_AST.getFirstChild().getNextSibling().getFirstChild().getFirstChild().getText();
			}
			else
			{
			    //  A transformation rule.
				return m_AST.getFirstChild().getNextSibling().getText();
			}
        }

        /**
         *  @return the nonterminal produced by this reduction.
         */
        @Override
        public String getGoalState()
        {
            return m_AST.getFirstChild().getText();
        }

        /**
         *  Set this rule's associated action code.
         */
        public void setReduceAction(JBurgReduceAction reduceAction)
        {
            this.reduceAction = reduceAction;
        }

        /**
         *  @return true if this node has no children (a "leaf" node).
         */
        public boolean isTerminalPattern()
        {
            return this.isFixedArity() && this.patternMatcher.getNominalArity() == 0;
        }

        /**
         *  @return true if this rule's pattern has a fixed number of "operand" subtrees.
         */
        public boolean isFixedArity()
        {
            return this.patternMatcher != null && !this.patternMatcher.hasNaryTail();
        }

        /**
         *  @return true if this rule needs an out-of-line method to check its cost.
         */
        public boolean needsCostFunction()
        {
            if ( this.isTerminalPattern() )
                return false;
            else if ( this.patternMatcher.hasNaryTail() )
                return  this.patternMatcher.getNominalArity() != this.patternMatcher.getMinimumNaryChildCount();
            else
                return true;
        }

        /**
         *  Is this production's cost constant in the
         *  context where it's to be invoked?
         *  @param productions - the set of productions at the
         *    point where this productions's cost is required.
         */
        @Override
        public boolean computesConstantCost(Multimap<String, JBurgProduction> productions)
        {
            return hasConstantCost();
        }

        /**
         *  Get the constant cost of a production.
         *  @param productions - the set of productions at the
         *    point where this production's cost is required.
         *  @return the cost, with overflow-safe addition.
         *  @pre computesConstantCost(productions) must be true.
         */
        @Override
        public int getConstantCost(Multimap<String, JBurgProduction> productions)
        {
            return getConstantCost();
        }

        @Override
        public String toString()
        {
            return m_AST.toStringTree();
        }

        /**
         * @return this rule's reduction logic
         * as platform-specific code.
         */
        public String getReductionCode()
        {
            if ( m_AST.getType() == SIMPLE_TRANSFORMATION_RULE )
            {
                // TODO: Turn this into a string template.
                return String.format(
                    "%s%s",
                    codeEmitter.genReturnValue(
                        codeEmitter.genPopFromStack(reducedValuesName, getReturnType(this.getGoalState()))
                    ),
                    codeEmitter.genEndStmt()
                );
            }
            else if ( hasASTOfType(m_AST, REDUCTION_ACTION) )
            {
                AST reduction = getASTByType(m_AST, REDUCTION_ACTION);
                return getCodeBlock(reduction);
            }
            else if ( hasASTOfType(m_AST, EXPLICIT_REDUCTION ) )
            {
                AST reduction = getASTByType(m_AST, EXPLICIT_REDUCTION);
                // TODO: String template.
                return "return " + emitExpression(getASTByType(reduction, PROCEDURE_CALL)) + ";";
            }
            else
            {
                throw new IllegalStateException("A reduction must specify an implementation block or callback function");
            }
        }
    }

    /**
     *  JBurgReduceAction holds action code fragments and their associated "parameters."
     *  A rule specifies "parameters" by naming individual subgoal states within a pattern
     *  rule specifiction, for example,
     *  <xmp>integer = PLUS ( integer i1, integer i2 )</xmp>
     *  In this example, i1 and i2 are the action routine's "parameters."  Since the action
     *  routines all share a common signature, the parameters are passed via the reducer's
     *  reduced values stack.
     */
    private class JBurgReduceAction
    {
        /** The reduction that triggers this action. */
        private final JBurgRule rule;
        
        /**
         * The non-terminal state produced by this action, e.g., expression from
         * <xmp>expression = ADD(expression l, expression r)</xmp>
         */
        private final String m_state;

        /**
         *  The antecedent reduction of a nonterminal-to-nonterminal rule.
         */
        private String antecedentState;
        
        /**
         *  The operator ID of a pattern rule; used to find content-handling
         *  code from the InodeAdapter2.
         */
        private String m_operator;
        
        /**
         * Names and types of the routine's parameters.
         * Types are given as their non-terminal state names; toString()
         * translates these BURM-centric types into the corresponding
         * types in the target language using the mapping set up by
         * the ReturnType directives in the input specification.
         */
        private Vector<ParameterDescriptor> m_parameterList = new Vector<ParameterDescriptor>();

        /**
         *  This action routine's index, assigned in entry order.
         *  The action routines are enumerated as action_1(JBurgNode p), action_2(JBurgNode p),
         *  etc. in the generated reducer.
         */
        int index;
        
        /**
         * Track name-to-subtree mappings to be emitted.
         */
        class NamedSubtree
        {
            String path;
            String name;
            
            NamedSubtree(String path, String name)
            {
                this.path = path;
                this.name = name;
            }
        }
        /**
         *  Saved name-to-subtree mappings.
         */
        Vector<NamedSubtree> namedChildNodes = new Vector<NamedSubtree>();
        
        /**
         * Construct a reduce action.
         * @param strActionCode - the action's implementation
         * code, in a target-specific language 
         */
        public JBurgReduceAction(JBurgRule rule, String resultState)
        {
            this.rule  = rule;
            this.m_state = resultState;
        }

        /**
         * Add a parameter to the action's parameter list.
         */
        public void addParameter(String parmName, String parmState, ParameterDescriptor.ArityType arityType)
            throws Exception
        {
            m_parameterList.add(new ParameterDescriptor(parmName, parmState, arityType));
        }

        /**
         *  Return this action's index (the N constituent of its generated name, action_N).
         */
        public int getIndex()
        {
            return this.index;
        }

        /**
         *  Set this action's index (the N constituent of its generated name, action_N).
         */
        public void setIndex(int index)
        {
            this.index = index;
        }

        /**
         * @return the non-terminal state this reduction derives.
         */
        public String getState()
        {
            return this.m_state;
        }

        /**
         *  @param operator the operator at the root of 
         *    the matched subtree.
         */
        public void setOperator(String operator)
        {
            this.m_operator = operator;
        }

        /**
         * @return the operator at the root of 
         *    the matched subtree.
         */
        public String getOperator()
        {
            return m_operator;
        }

        /**
         *  @return the reduction action's code, with
         *  prologue logic to pop its parameters off the stack.
         */
        public String emit()
        {
            String result = "";

            for ( ParameterDescriptor next_param: m_parameterList )
            {
                String paramName = next_param.paramName;
                String paramType = getReturnType(next_param.paramState);

				if ( next_param.arityType == ParameterDescriptor.ArityType.Variable )
				{
					paramType = codeEmitter.genNaryContainerType(paramType);
				}
                
                result += codeEmitter.genActionRoutineParameter(reducedValuesName, paramType, paramName);
            }
            
            for ( NamedSubtree named_child: this.namedChildNodes )
            {
                result += "\n\t" + codeEmitter.genLocalVar(iNodeClass, named_child.name, named_child.path) + ";";
            }

            // insert 2 tabs into each line, and convert \r\n to \n so the generated
			// files have consistent line endings.
			String[] actionlines = this.rule.getReductionCode().split("\n");
			for( int l = 0; l < actionlines.length; ++l )
			{
				if(actionlines[l].endsWith("\r"))
				{
					actionlines[l] = actionlines[l].substring(0,actionlines[l].length()-1);
				}
				result += "\n\t\t" + actionlines[l];
			}

			return result;
        }

        /**
         *  Track a name-to-subtree mapping.  These are unreduced I-nodes,
         *  typically a terminal identified by pattern match.
         *  @param path - the path from the root node to the subtree.
         *  @param name - the name to give to the subtree.
         */
        public void addNamedSubtree(String path, String name)
        {
            namedChildNodes.add(new NamedSubtree(path, name));
        }

        /**
         *  Set this reduction's antecedent state, which
         *  determines what reduction needs to run before
         *  this one to transform one nonterminal to another.
         */
        public void setAntecedentState(String antecedentState)
        {
            this.antecedentState = antecedentState;
        }

        /**
         *  @return true if this rule has an antecedent state,
         *    which also means it's a nonterminal-to-nonterminal rule.
         */
        public boolean hasAntecedentState()
        {
            return getAntecedentState() != null;
        }

        /**
         *  @return this rule's antecedent state,
         *    or null if no antecedent is present.
         */
        public String getAntecedentState()
        {
            return this.antecedentState;
        }
    }

    /**
     *  ClosureRecord tracks the target state, cost, and action code associated
     *  with a reduction from an antecedent nonterminal to a target nonterminal state.
     */
    private class ClosureRecord implements JBurgProduction
    {
        private JBurgRule rule;

        public ClosureRecord(JBurgRule rule) throws Exception
        {
            this.rule = rule;
        }

        /**
         *  Get the closure's uncached cost.
         *  @deprecated Will be removed when rules start caching costs.
         */
        public String getCost(String baseNT)
        {
            return this.rule.getCost(baseNT);
        }

        /**
         *  Get the closure's (potentially) cached cost.
         */
        @Override
        public String getCachedCost()
        {
            return getCachedCost(reducerNodeName);
        }

        public String getCachedCost(String nodeName)
        {
            String cost = this.rule.getCost(nodeName);

            //  The cost function result will be cached;
            //  return a unique accessor method name.
            if ( hasCostFunction() )
                return JBurgGenerator.this.codeEmitter.genCallMethod(null, String.format("getCostFunctionResult_%h", cost));
            else
                return cost;
        }

        /**
         *  @return the rule's hasCostFunction() result.
         */
        public boolean hasCostFunction()
        {
            return rule.hasCostFunction();
        }

        /**
         *  @return the rule's getConstantCost() result.
         */
        public int getConstantCost()
        {
            return this.rule.getConstantCost();
        }

        /**
         *  @return the rule's hasConstantCost() result.
         */
        public boolean hasConstantCost()
        {
            return this.rule.hasConstantCost();
        }

        /**
         *  @return the rule's getAntecedentState() result.
         */
        public String getAntecedentState()
        {
            return this.rule.getAntecedentState();
        }

        /**
         *  @return the rule's getGoalState() result.
         */
        @Override
        public String getGoalState()
        {
            return this.rule.getGoalState();
        }

        /**
         *  @return this closure's reduce action.
         */
        @Override
        public JBurgReduceAction getReduceAction()
        {
            return rule.getReduceAction();
        }

        /**
         *  Closure records are equal if they have the same goal state, action, and cost.
         *  @see java.lang.Object#equals
         *
         * @param o -- the object to test for equality.
         * @return true if o is a ClosureRecord and it equals this one.
         *
         */
        @Override
        public boolean equals(Object o)
        {
            boolean result = false;

            if (o instanceof ClosureRecord)
            {
                ClosureRecord cuz = (ClosureRecord) o;

                result = getGoalState().equals(cuz.getGoalState());
                result &= getReduceAction().equals(cuz.getReduceAction());
                result &= this.rule.getCost(reducerNodeName).equals(cuz.rule.getCost(reducerNodeName));
            }

            return result;
        }

        /**
         *  @return true if the cost of this closure is known to be zero.
         */
        public boolean costIsZero()
        {
            return hasConstantCost() && getConstantCost() == 0;
        }

        /**
         *  @return the computation of this closure's cost,
         *    which is somewhat error-prone when open coded.
         */
        public String getCostComputation()
        {
            return getCostComputation(
                null, 
                JBurgGenerator.this.codeEmitter.genCallMethod(
                    null,
                    "getCost",
                    getNonterminal(this.getAntecedentState())
                )
            );
        }

        /**
         * @param stem the stem of the cached cost computation.
         * @param antecedentCost the caller's computation of this
         * closure's antcedent cost.
         * @return
         */
        public String getCostComputation(String stem, String antecedentCost)
        {
            if ( this.costIsZero() )
            {
                return antecedentCost;
            }
            else
                return codeEmitter.genOverflowSafeAdd(this.getCachedCost(), antecedentCost);
        }

        /**
         *  Precompute the cost of this closure if possible.
         *  @param productions - the set of productions at the
         *    point where the closure's cost is required.  If
         *    the closure derives a single pattern-match 
         *    rule that has a constant cost, and the closure
         *    itself has a constant cost, then the cost
         *    can be precomputed.
         *  @return the precomputed cost, or the result of
         *    calling {@link getCostComputation()} to get
         *    a non-constant cost.
         */
        public String getCostComputation(Multimap<String, JBurgProduction> productions)
        {
            //  A closure back to a pattern-match with constant cost can be precomputed.
            if ( this.computesConstantCost(productions) )
            {
                return Integer.toString(getConstantCost(productions));
            }
            else
            {
                //  Return the naive cost computation.
                return getCostComputation();
            }
        }

        /**
         *  Get the constant cost of a closure.
         *  @param productions - the set of productions at the
         *    point where the closure's cost is required.
         *  @return the cost, with overflow-safe addition.
         *  @pre computesConstantCost(productions) must be true.
         */
        @Override
        public int getConstantCost(Multimap<String, JBurgProduction> productions)
        {
            long constantAntecedent;
            JBurgProduction production = productions.get(this.getAntecedentState()).get(0);

            if ( production instanceof JBurgRule )
                constantAntecedent = ((JBurgRule)production).getConstantCost();
            else
                constantAntecedent = ((ClosureRecord)production).getConstantCost(productions);

            //  Integer-overflow safe addition.
            @SuppressWarnings("cast")
            long accum = (long)this.getConstantCost() + constantAntecedent;

            if ( accum < Integer.MAX_VALUE )
                return (int) accum;
            else
                return Integer.MAX_VALUE;
        }

        /**
         *  Does this closure compute a constant cost
         *  in the context of the given productions?
         *  @param productions - the set of productions at the
         *    point where the closure's cost is required.  If
         *    the closure derives a single pattern-match 
         *    rule that has a constant cost, and the closure
         *    itself has a constant cost, then the cost
         *    can be precomputed.
         *  @return true if the computed cost is constant.
         */
        @Override
        public boolean computesConstantCost(Multimap<String, JBurgProduction> productions)
        {
            boolean result = false;
            if ( this.hasConstantCost() )
            {
                ArrayList<JBurgProduction> antecedents = productions.get(this.getAntecedentState());

                if ( antecedents.size() == 1 )
                {
                    JBurgProduction production = antecedents.get(0);

                    if ( production instanceof JBurgRule )
                        result = ((JBurgRule)production).hasConstantCost();
                    else
                        result = ((ClosureRecord)production).computesConstantCost(productions);
                }
            }

            return result;
        }

        public String toString()
        {
            return String.format("Closure{%s=%s:%s}", getGoalState(), getAntecedentState(), getCostComputation());
        }
    }

	/**
	 *  A ParameterDescriptor is a pair (name, subgoalState) which parallels
	 *  the FOO(name, subgoalState) found in the specification's syntax.
	 */
    private static class ParameterDescriptor
    {
        String paramName;
        String paramState;
        ArityType arityType;
        
        enum ArityType {Fixed, Variable}

        ParameterDescriptor(String paramName, String paramState, ArityType arityType)
        {
            this.paramName = paramName;
            this.paramState = paramState;
            this.arityType = arityType;
        }
    }

    /**
     *  Finish semantic analysis of the pattern rules.
     */
	private void ruleSemanticAnalysis() throws Exception
    {
        for (Vector<JBurgRule> vPatternRules: this.patternRules.values())
        {
            for (JBurgRule rule: vPatternRules)
            {
                for (JBurgPatternMatcher subgoal: rule.patternMatcher.getParameterizedSubtrees())
                {
                    rule.getReduceAction().addParameter(
                            subgoal.getParameterName(), 
                            subgoal.getSubgoal(), 
                            subgoal.isNary()? ParameterDescriptor.ArityType.Variable: ParameterDescriptor.ArityType.Fixed
                            );
                }

                //  Add the named subtrees as a convenience.
                for ( JBurgPatternMatcher named_terminal: rule.patternMatcher.getNamedSubtrees())
                {
                    rule.getReduceAction().addNamedSubtree(
                            named_terminal.generateReduceTimePath(codeEmitter, reducerNodeName, this.iNodeAdapter, this.adapter2 != null), 
                            named_terminal.getParameterName()
                            );
                }

                //  If the pattern matches an operator, send that operator
                //  to the rule so it can be matched in content-access snippets.
                if ( rule.patternMatcher.matchesOperator() )
                    rule.getReduceAction().setOperator(rule.patternMatcher.getOperator());
            }
        }
    }

    /**
     * If the target does not support specialized annotations,
     * emit a generalized (and slow) analysis routine.
     */
	private void emitComputeCostMatrixFunction(String iNodeClass, PrintStream output) throws Exception 
	{
		//  Emit the method declaration and local variables.
		genDeclareMethod( output, Modifier.PRIVATE, "void", "computeCostMatrix", "JBurgAnnotation", "node");
		genBeginBlock(output);
		genLocalVar(output,  "long", "iCost", null );

		//  The calculations are based on finding patterns
		//  that can label the node's type.
		genSwitch(output, this.codeEmitter.genCallMethod( "node", "getOperator" ));

		//  Try each pattern for a match.
		//  Any pattern that matches then checks its 
		//  cost to see if it's less than the current
		//  best cost for the node, and replaces the 
		//  node's current rule and cost with its own
		//  if it's the new best cost.

		for (Vector<JBurgRule> vPatternRules: this.patternRules.values())
		{
            genCase( output, vPatternRules.firstElement().getOperator() );
            
            //  Group rules by the patterns they match.
            Multimap<String, JBurgRule> rules_by_pattern = new Multimap<String, JBurgRule>();

            for (JBurgRule rule: vPatternRules)
            {
                String encoded_pattern = rule.getEncodedPattern();
                rules_by_pattern.addToSet(encoded_pattern, rule);
            }

            
            for ( String encoded_pattern: rules_by_pattern.keySet() )
            {
                ArrayList<JBurgRule> current_rules = rules_by_pattern.get(encoded_pattern);

                //  Emit the structural pattern match if it was not optimized out.
                if ( !encoded_pattern.equals(nilPattern))
                    genIf(output, encoded_pattern);
                
                //  Begin a block whether there was a test or not, to scope variables.
                //  TODO: This will not work for a target that doesn't block-scope locals.
                genBeginBlock(output);
                
                //  Find common path and cost subexpressions.
                Multimap<JBurgPatternMatcher, JBurgPatternMatcher> aggregated_matchers = new Multimap<JBurgPatternMatcher, JBurgPatternMatcher>();
                
                Multimap<String, JBurgPatternMatcher> costs = new Multimap<String, JBurgPatternMatcher>();
                
                for ( JBurgRule rule: rules_by_pattern.get(encoded_pattern))
                {
                    for (JBurgPatternMatcher subgoal: rule.patternMatcher.getParameterizedSubtrees() )
                    {
                        subgoal.aggregatePaths(aggregated_matchers);
                        costs.addToSet(subgoal.generateCost(codeEmitter, "node"), subgoal);
                    }
                }
                
                //  Factor out common paths.
                int factored_path_count = 0;
                for ( JBurgPatternMatcher key: aggregated_matchers.keySet())
                {
                    ArrayList<JBurgPatternMatcher> matchers = aggregated_matchers.get(key);
                    
                    if ( matchers.size() > aggregatePathThreshold  )
                    {
                        String factored_path_var = "factored_path_" + Integer.toString(++factored_path_count);
                        genLocalVar(output, "JBurgAnnotation ", factored_path_var, key.generatePathToRoot(codeEmitter, "node"));
                        
                        for ( JBurgPatternMatcher matcher: matchers)
                            matcher.factoredPath = factored_path_var;
                        
                    }
                }
                
                //  Now that the paths have been factored, factor out common costs.
                int factored_cost_count = 0;
                
                for (String key: costs.keySet() )
                {
                    if ( costs.get(key).size() > 1 )
                    {
                        String factored_cost_var = "factored_cost" + Integer.toString(++factored_cost_count);
                        //  Regenerate the cost because it may now use factored path expressions.
                        ArrayList<JBurgPatternMatcher> matchers = costs.get(key);
                        genLocalVar(output, "int", factored_cost_var, matchers.get(0).generateCost(codeEmitter, "node"));
                        for ( JBurgPatternMatcher subgoal: matchers)
                        {
                            subgoal.factoredCost = factored_cost_var;
                        }
                    }
                }

                //  Emit the cost checks and their labeling logic.
                
                int patternPosition = 0;

                for ( JBurgRule rule: current_rules )
                    emitPatternRule(rule, iNodeClass, output, patternPosition++ == 0);

                genEndBlock(output);
            }

            genEndCase(output);
		}

		genEndSwitch(output);
		genEndBlock(output);
	}
	
	/**
	 *  Emit a routine that inspects the operators found at compiler compile time,
	 *  and emits code to get the corresponding node's Nth child.  This routine is
     *  only called if the I-node adapter has special logic to emit for getNthChild(idx).
	 *  @param output - the current output stream.
	 */
	private void emitGetNthChild(PrintStream output)
	{
	    genDeclareMethod(
            output,
            Modifier.PRIVATE,
            this.iNodeClass,
            "getNthChild",
	        genFormals(this.iNodeClass, "node", "int", "index"),
            throwsNothing
        );
	    genBeginBlock(output);
	    genLocalVar(output, this.iNodeClass, "result", null);
	    
	    genSwitch(output, this.iNodeAdapter.genGetOperator("node", this.codeEmitter));
	    for ( String opcode: this.allOperators)
	    {
	        Integer choice_count = this.adapter2.getMaxNthChildChoice(opcode);
	        if ( choice_count != null )
	        {
                genCase(output, opcode);
	            genSwitch(output, "index");
	            for ( int i = 0; i < choice_count; i++ )
	            {
	                genCase(output, Integer.toString(i));
	                genAssignment(output, "result", this.adapter2.genGetNthChild(opcode, "node", i, codeEmitter));
	                genEndCase(output);
	                
	            }
                //  generate the default case for the index.
                {
                    genDefaultCase(output);
                    genAssignment(output, "result", this.adapter2.genGetDefaultChild(opcode, "node", "index", codeEmitter));
                    genEndCase(output);
                }

                genEndSwitch(output);
	            genEndCase(output);
	        }
	        
	    }
        //  generate the default case for the opcode.
        {
            genDefaultCase(output);
            genAssignment(output, "result", this.iNodeAdapter.genGetNthChild("node", "index", codeEmitter));
            genEndCase(output);
        }
	    genEndSwitch(output);
	    genReturnValue(output, "result");
	    genEndBlock(output);
	}

    /**
     * Emit a pattern matcher that does its dynamic programming at compile time.
     */
	private void emitPatternRule(JBurgRule p, String iNodeClass, PrintStream output, boolean isFirstRule ) throws Exception
	{
	    genComment(output, "Try matching " + p.getOperator() + " ==> " + p.getGoalState());
	    
        /*
		 *   Emit code to compute this rule's cost, which is the cost of the rule itself 
		 *   and all the non-terminals associated with it; then emit a comparison with
		 *   the current best cost.
		 */
		Vector<String> sub_costs = new Vector<String>();

		//  Compute the basic cost of the pattern match.
		sub_costs.add(
		    p.getCost( 
    			codeEmitter.genCast( 
    				iNodeClass, 
    				codeEmitter.genAccessMember("node", "m_node")
    			)
            )
		);
		
	   //  Try and compute a constant cost.
        Integer constant_cost = null;

		if ( p.hasConstantCost() )
		{
		    constant_cost = p.getConstantCost();
		}

		
		for ( JBurgPatternMatcher subgoal: p.patternMatcher.getParameterizedSubtrees() )
		{
		    String subgoal_cost = subgoal.generateCost(codeEmitter, "node");
		    if ( subgoal_cost != null )
		    {
		        sub_costs.add(subgoal_cost);
		        constant_cost = null;
		    }
		}
		
		String cost_factor;
        boolean checkCostFactor = true;
		
		if ( constant_cost != null )
		{
		    cost_factor = constant_cost.toString();
            checkCostFactor = constant_cost == Integer.MAX_VALUE;
		}
		else if ( sub_costs.size() == 1 )
		{
		    cost_factor = sub_costs.get(0);
		}
		else
		{
            //  Compute the cost using a long accumulator.
		    cost_factor = "iCost";

    		String match_cost = codeEmitter.genOverflowSafeAdd(sub_costs.get(0), sub_costs.get(1));
        		
            for ( int i = 2; i < sub_costs.size(); i++ )
            {
                assert sub_costs.get(i) != null;
                match_cost = codeEmitter.genOverflowSafeAdd(sub_costs.get(i), match_cost);
            }
    		
    		genAssignment(output, "iCost", match_cost);
		}

        if ( !isFirstRule || checkCostFactor )
        {
            //  Does this pattern match cost less than the previously best match?
            String getCostCall = this.codeEmitter.genCallMethod ( "node", "getCost", codeEmitter.genGetGoalState(p) );
            String costCheck = codeEmitter.genCmpLess( getCostCall, cost_factor );
            genIf(output, costCheck );
        }

		output.print( codeEmitter.genBeginBlock() );

		/*
		 *   Emit code to handle a successful match.
		 */

		//  Reset the reduce-time rule to fire.
        String strRule = String.valueOf(p.getReduceAction().getIndex());
        
        //  Now that the comparison has been done using long arithmetic,
        //  iCost can be safely truncated to fit in an int.
        if ( cost_factor.equals("iCost"))
            cost_factor = codeEmitter.genCast("int", cost_factor);

        genExpressionStmt(
            output, 
			this.codeEmitter.genCallMethod(
				"node", 
				"reset", 
				codeEmitter.genGetGoalState(p), cost_factor, strRule
			)
		);

        if (hasClosure(p))
		{
            genExpressionStmt(
                output,
				this.codeEmitter.genCallMethod( 
					null, 
					"closure_" + p.getGoalState(),
					"node", cost_factor
				)
			);
        }

        //  End the block begun by the cost comparison above.
        genEndBlock(output);
    }
	
    /**
     * Translate a pattern specification into a pattern matcher.
     * @param pattern_root the root of the pattern.
     */
	private JBurgPatternMatcher generateMatcher(AST pattern_root)
	throws Exception
	{   
        JBurgPatternEncoder patternBURM = new JBurgPatternEncoder();

        //  As we traverse the subtree, we may find parameterized subtrees,
        //  for example, in ADD(expr lhs, expr rhs) lhs and rhs are paramterized
        //  subtrees.  These subtrees play several parts in the computation of the
        //  locally-optimal reduction:
        //  -  They contribute to the rule's computed cost.
        //  -  The reduction's action code may refer to these elements by name.
        //  -  If the rule is of the form OP(nttype1 a [, nttype2 b]), then the rule
        //     must enforce this reduce-time goal state on its subtrees' reductions.
        patternBURM.setSubgoals(new Vector<JBurgPatternMatcher>());
        
        //  There may also be named terminals in the pattern;
        //  as a convenience, record these so that they
        //  can be used by name in the reduction.
        patternBURM.setNamedterminals(new Vector<JBurgPatternMatcher>());
        
        patternBURM.setOperators(this.allOperators);

        try
        {
            patternBURM.burm( pattern_root );
        }
        catch (IllegalStateException burm_error )
        {
            if ( this.patternMatcherDumpFile != null )
            {
                //  Dump the BURM's debugging info.
                patternBURM.dump(new PrintWriter(this.patternMatcherDumpFile));
            }
            
            throw burm_error;
        }
        
        JBurgPatternMatcher recognizer = (JBurgPatternMatcher) patternBURM.getResult();
        recognizer.setParameterizedSubtrees(patternBURM.getSubgoals());
        recognizer.setNamedSubtrees(patternBURM.getNamedterminals());
        
        return recognizer;
	}

    /**
     * Emit code to call the action routines.
     */
	private void emitActions(ArrayList<JBurgReduceAction> reduceActions, String iNodeClass, PrintStream output)
	{
		int i = 1;

		//  Print out the individual action routines.
		for (JBurgReduceAction nextAction: reduceActions )
		{
            output.println();
			genComment(output, nextAction.getState());
			
            //  Compute the type of the i-node.
            String typeOfOperatorINode;
			final String actionOperator = nextAction.getOperator();

			if ( opcodeNodeTypes.containsKey(actionOperator) )
			    typeOfOperatorINode = opcodeNodeTypes.get(actionOperator);
			else
			    typeOfOperatorINode = iNodeClass;

			genDeclareMethod(
                output,
				Modifier.PRIVATE,
				getReturnType(nextAction.getState()),
				"action_" + String.valueOf(i++),
				genFormals(typeOfOperatorINode, reducerNodeName),
				throwsException
			);
				
			genBeginBlock(output);

			String action_routine = nextAction.emit();
			
			//  Replace #OPERATOR# with a content-access snippet.
			if ( this.adapter2 != null && nextAction.getOperator() != null )
			{
			    String content_access = this.adapter2.genGetContent(nextAction.getOperator(), this.reducerNodeName, nextAction.getState(), codeEmitter);
			    
			    if ( content_access != null )
			    {
			        String content_pattern = "#" + nextAction.getOperator() + "#";
			        action_routine = action_routine.replaceAll( content_pattern, content_access );
			    }
			}
			//  Replace #goalstate with the name of the action's input node.
			
			String goalPattern = "#" + nextAction.getState();
			action_routine = action_routine.replaceAll( goalPattern, this.reducerNodeName );
			
			output.print( action_routine );

			genEndBlock(output);
		}

		//  Emit their common dispatch routine.
		genDeclareMethod(
            output,
			Modifier.PRIVATE,
			"void",
			"dispatchAction",
			genFormals("JBurgAnnotation", "___node", "int", "iRule"),
			throwsException
		);

		genBeginBlock(output);

        genLocalVar(
            output,
            iNodeClass,
            this.reducerNodeName,
            this.codeEmitter.genCallMethod("___node", "getNode")
        );

		genSwitch(output, "iRule" );

		//  Emit the dispatch case statement.
		for (i = 1; i <= reduceActions.size(); i++)
		{
            JBurgReduceAction action = reduceActions.get(i-1);

			genCase(output, String.valueOf(i));
            {
                //  If this is a nonterminal-to-nonterminal
                //  transformation, run the antecedent
                //  reduction action.
                if ( action.hasAntecedentState() )
                {
                    genExpressionStmt(
                        output,
                        this.codeEmitter.genCallMethod(
                            "this",
                            "reduceAntecedent",
                            "___node", codeEmitter.genGetGoalState(action.getAntecedentState())
                        )
                    );
                }

                final String operatorName = action.getOperator();
                final String nodeTypeForOperator =
                    operatorName != null ? opcodeNodeTypes.get(operatorName) : null;
                String nodeParameterString = reducerNodeName;
                if (nodeTypeForOperator != null)
                    nodeParameterString = codeEmitter.genCast(nodeTypeForOperator, nodeParameterString);
                
                genExpressionStmt(
                    output,
                    codeEmitter.genPushToStack(
                        reducedValuesName,
                        this.codeEmitter.genCallMethod(
                            "this",
                            "action_" + String.valueOf(i),
                            nodeParameterString
                        )
                    )
                );
            }
			genEndCase(output);
		}

		genDefaultCase(output);
        {
            genThrow(output, "\"Unmatched reduce action \" + iRule" );
        }
        genEndBlock(output); // genEndCase() without unreachable break
	    
		genEndSwitch(output);
		genEndBlock(output);
	}

    /**
     * Emit the state identifiers.
     */
	private void emitNTConstants(Set<String> goal_states, PrintStream output)
	{
        // Don't emit these constants if the nonterminal type
        // was explicitly set.
        if ( ! this.ntType.equals("int") )
            return;

		Iterator<String> keys = goal_states.iterator();

		int nthConstant = 0;

		output.print(codeEmitter.genBeginLine());
		while (keys.hasNext())
		{
			nthConstant++;

			String strNTName = codeEmitter.genGetGoalState(keys.next().toString() );

            genInstanceField(
                output, 
                Modifier.PUBLIC + Modifier.STATIC + Modifier.FINAL,
                "int",
                strNTName,
                String.valueOf(nthConstant)
			);
		}
	
        genInstanceField(
            output,
            Modifier.PUBLIC + Modifier.STATIC + Modifier.FINAL,
            "int",
            "nStates",
            String.valueOf(nthConstant)
		);
	}

    /**
     * Emit cost functions.
     */
	public void emitCostFunctions(ArrayList<AST> costFunctions, String iNodeClass, PrintStream output) 
	{
		String[][] costParms = genFormals( iNodeClass, reducerNodeName );

		for (int i = 0; i < costFunctions.size(); i++)
		{
			AST currentNode = costFunctions.get(i);

			String functionName = currentNode.getFirstChild().getText();

			genDeclareMethod(output, Modifier.PRIVATE, "int", functionName, costParms, throwsNothing );
			output.print( codeEmitter.genBeginLine() );
			output.print( getCodeBlock(currentNode) );
		}
	}

	public void emitClosures(Map<String, Vector<ClosureRecord>> closureSets, String iNodeClass, PrintStream output)
	{
        //  TODO: This method can be removed once the C++ target
        //  has been updated to 1.9.0 style semantics.
		for (String strClosureNT: closureSets.keySet())
		{
			genDeclareMethod(
				output,
                Modifier.PRIVATE,
                "void",
                "closure_" + strClosureNT,
                genFormals("JBurgAnnotation", reducerNodeName, "long", "c"),
                throwsNothing
			);

			genBeginBlock(output);

			genLocalVar(output, "long", "iCost", null );

			for (ClosureRecord newClosure: closureSets.get(strClosureNT) )
			{
				String strRule;

				if (newClosure.getReduceAction() != null)
					strRule = String.valueOf(newClosure.getReduceAction()
													   .getIndex());
				else
					strRule = "0";

				String closureCost = newClosure.getCost( codeEmitter.genAccessMember(reducerNodeName, "m_node"));

				if (closureCost.equals("0") )
				{
					genAssignment( output, "iCost", "c");
				}
				else
				{
					genAssignment( output, "iCost",  codeEmitter.genOverflowSafeAdd( "c", closureCost ));
				}

                genIf(
                    output,
                    codeEmitter.genCmpLess(
                        this.codeEmitter.genCallMethod (
                            reducerNodeName,
                            "getCost", 
                            getNonterminal(newClosure.getGoalState())
                        ),
                        "iCost"
					)
				);
				output.print( codeEmitter.genBeginBlock() );

                genExpressionStmt(
                    output,
					this.codeEmitter.genCallMethod(
						reducerNodeName,
						"reset",
						getNonterminal(newClosure.getGoalState()), codeEmitter.genCast("int","iCost"), strRule
					)
				);

                genExpressionStmt(
                    output,
					this.codeEmitter.genCallMethod(
						reducerNodeName,
						"recordAntecedent",
						getNonterminal(newClosure.getGoalState() ), getNonterminal(strClosureNT )
					)
				);

				if (hasClosure(newClosure))
				{
                    genExpressionStmt(
                        output,
						this.codeEmitter.genCallMethod(
							"this",
							"closure_" + newClosure.getGoalState(),
							reducerNodeName, "iCost"
						)
					);
				}

				genEndBlock(output);
			}

			genEndBlock(output);
		}
	}

    /**
     * Emit the label(ast) function.
     */
	public void emitLabelFunction(PrintStream output)
	{
        StringTemplate labelMethod = getTemplate(
            "method",
            "decl", codeEmitter.declareMethod(
                Modifier.PUBLIC,
                "JBurgAnnotation",
                "label",
                genFormals(this.iNodeClass, JBurgGenerator.initialParamName),
                throwsNothing
            )
        );

        addVariableDecl(labelMethod, codeEmitter.getAnnotationType(), "result", codeEmitter.genNullPointer());
        addVariableDecl(labelMethod, "int", "i", "0");
        addVariableDecl(
            labelMethod,
            "int",
            "arity",
            this.iNodeAdapter.genGetArity(JBurgGenerator.initialParamName, codeEmitter)
        );

        addAssignment(
            labelMethod,
            "result",
            codeEmitter.supportsSpecializedAnnotations()?
                this.codeEmitter.genCallMethod( "this", "getJBurgAnnotation", JBurgGenerator.initialParamName):
                this.codeEmitter.genNewObject( "JBurgAnnotation", JBurgGenerator.initialParamName, "nStates + 1" )
        );

		//  Generate a loop over children that labels them,
		//  and adds their labelled JBurgAnnotation nodes to the current node.

        StringTemplate whileLoop = getTemplate(
            "whileStatement",
            "test", codeEmitter.genCmpLess("arity", "i")
        );

        addContents(labelMethod, whileLoop);

        addExprStatement(
            whileLoop,
            this.codeEmitter.genCallMethod(
                "result",
                "addChild",
                genCallLabel("i")
            )
        );
            
        addAssignment(whileLoop, "i", codeEmitter.genAddition("i", "1") );

		//  Call the costing function.
        if ( ! this.codeEmitter.supportsSpecializedAnnotations() )
            addExprStatement(
                labelMethod,
                this.codeEmitter.genCallMethod( "this", "computeCostMatrix", "result")
            );

		addReturnValue(labelMethod, "result" );
		output.println(labelMethod);
	}

    /**
     *  Emit the subclasses of JBurgAnnotation that encode data for specific pattern matches.
     *  @param output - the destination output stream.
     */
    void emitCompressedAnnotations(PrintStream output)
    {
        //  Emit an annotation object for each equivalence class of rules.
        for ( String operator: this.compressedAnnotations.getOperators() )
            for ( ArrayList<JBurgRule> currentRules : this.compressedAnnotations.getRulesFor(operator) )
                emitAnnotation(output, currentRules);

        //  Emit getJBurgAnnotation
        StringTemplate getJBurgAnnotation = getTemplate(
            "method",
            "decl", codeEmitter.declareMethod(
                Modifier.PUBLIC,
                "JBurgAnnotation",
                "getJBurgAnnotation",
                genFormals(this.iNodeClass, "node"),
                throwsNothing
            )
        );

        // A code fragment that fetches the node's actual arity.
        String actualArity = this.iNodeAdapter.genGetArity("node", codeEmitter);

        StringTemplate switchOnNode = getTemplate(
            "switch",
            "expr", this.iNodeAdapter.genGetOperator("node", this.codeEmitter)
        );

        addContents(getJBurgAnnotation, switchOnNode);

        for ( String operator: compressedAnnotations.getOperators() )
        {
            StringTemplate opCase = getTemplate("switchCase", "label", operator);
            switchOnNode.setAttribute("contents", opCase);

            for ( ArrayList<JBurgRule> rules : this.compressedAnnotations.getRulesFor(operator) )
            {
                int arity = getMinumumArity(rules);

                StringTemplate ifStmt = getTemplate(
                    "ifThen",
                    "condition", hasNaryness(rules)?
                        getTemplate("greaterThanOrEquals", "lhs", actualArity, "rhs", arity):
                        getTemplate("equals", "lhs", actualArity, "rhs", arity),
                    "contents", getTemplate(
                        "returnStmt",
                        "value", getTemplate(
                            "newObject",
                            "type", getSpecializedClassName(rules),
                            "parameters", "node"
                        )
                    )
                );
                opCase.setAttribute("contents", ifStmt);
            }
        }

        getJBurgAnnotation.setAttribute(
            "contents",
            getTemplate(
                "returnStmt",
                "value", codeEmitter.genNewObject( "PlaceholderAnnotation", "node", actualArity)
            )
        );
        output.println(getJBurgAnnotation);
    }

    /**
     *  Emit an annotation subclass.
     *  @param output - the output stream.
     *  @param currentRules - the set of rules that are this annotation's
     *    domain.  The rules all have the same operator, but their arity
     *    may differ if some of them are n-ary.
     */
    void emitAnnotation(PrintStream output, ArrayList<JBurgRule> currentRules )
    {
        int nominalArity = getMinumumArity(currentRules);

        // The coalesced graph of closures for all the rules.
        ClosureGraph closureCosts = new ClosureGraph();

        //  Factored common subtrees of each pattern matcher.
        //  TODO: These can be further coalesced by using the right comparator.
        Multimap<JBurgPatternMatcher,JBurgPatternMatcher> commonSubtrees = new Multimap<JBurgPatternMatcher,JBurgPatternMatcher>();

        // Rules only by nonterminal.
        Multimap<String, JBurgProduction> patternMatches = new Multimap<String, JBurgProduction>();
 
        //  Rules and closures by nonterminal.
        Multimap<String, JBurgProduction> productions = new Multimap<String, JBurgProduction>();

        //  Cached subtree costs.
        Map<String, String> cachedCosts = new HashMap<String, String>();

        //  Are all these rules fixed-arity?
        //  If so, the annotation knows its arity a priori.
        boolean fixedArity = !hasNaryness(currentRules);

        /*
         *  **  Semantic analysis  **
         */
        
        //  Populate the closure graph and factor the pattern matchers.
        for ( JBurgRule rule: currentRules)
        {
            patternMatches.addToSet(rule.getGoalState(), rule);
            productions.addToSet(rule.getGoalState(), rule);

            closureCosts.addClosures(rule.getGoalState());

            if ( fixedArity )
                rule.patternMatcher.setFixedArityContext(true);

            for ( JBurgPatternMatcher subgoal: rule.patternMatcher.getParameterizedSubtrees() )
                subgoal.findFactors(commonSubtrees);

            if ( rule.patternMatcher.getNominalArity() > 0 || rule.patternMatcher.hasNaryness() )
            {
                cachedCosts.put(rule.getGoalState(), String.format("cachedCostFor_%s", rule.getGoalState()));
            }
        }

        //  Add the closures from the coalesced closure graph
        //  to the set of productions.
        for ( Map.Entry<String, ArrayList<ClosureRecord>> closures: closureCosts.entrySet() )
        {
            for ( ClosureRecord closure: closures.getValue() )
                productions.addToSet(closures.getKey(), closure);
        }

        //  Emitting these variable declarations mutates the 
        //  pattern matchers, so they can only be emitted once.
        //  Capture the results so we can write it more than once.
        Map<JBurgPatternMatcher,String> factored_variables = emitFactoredPathVariables(commonSubtrees);

        /*
         * Begin code-gen of the class.
         */
        output.println();
        String annotationClassName = getSpecializedClassName(currentRules);
        StringTemplate annotationClass = getTemplate(
            "innerClass",
            "className", annotationClassName,
            "baseClassName", "JBurgSpecializedAnnotation"
        );

        this.codeEmitter.setDefaultAttribute("className", annotationClassName);
        
        //  Emit a field for each child.
        for ( int fieldIdx = 0; fieldIdx < nominalArity; fieldIdx++ )
        {
            addContents(
                annotationClass,
                codeEmitter.genInstanceField( Modifier.PRIVATE, "JBurgAnnotation", String.format("subtree%d", fieldIdx), null)
            );
        }

        for ( String cachedCostVar: cachedCosts.values() )
        {
            addContents(
                annotationClass,
                codeEmitter.genInstanceField( Modifier.PRIVATE, "int", cachedCostVar, UNINITIALIZED)
            );
        }

        if ( !fixedArity )
        {
            addContents(
                annotationClass,
                codeEmitter.genInstanceField( 
                    Modifier.PRIVATE,
                    codeEmitter.genNaryContainerType("JBurgAnnotation"),
                    "narySubtrees",
                    getTemplate(
                        "newObject",
                        "type", codeEmitter.genNaryContainerType("JBurgAnnotation")
                    ).toString()
                )
            );
        }

        // Emit the constructor.
        StringTemplate constructor = getTemplate(
            "method",
            "decl", codeEmitter.declareMethod(
                Modifier.PUBLIC,
                "",
                annotationClassName,
                genFormals(this.iNodeClass, "node"),
                throwsNothing
            )
        );
        addContents(annotationClass, constructor);

        addContents(
            constructor,
            getTemplate(
                "callSuper",
                "parameters", "node"
            )
        );

        // getPatternMatchCost(nonterminal)
        if ( patternMatches.keySet().size() > 0 )
        {
            StringTemplate getPatternMatchCost = getTemplate(
                "method",
                "decl", codeEmitter.declareMethod(
                    Modifier.PRIVATE,
                    "int",
                    "getPatternMatchCost",
                    genFormals(this.ntType, "goalState"),
                    throwsNothing
                )
            );

            addContents(annotationClass, getPatternMatchCost);
            addVariableDecl(getPatternMatchCost, "int", "result", codeEmitter.genMaxIntValue());

            StringTemplate switchOnGoal = getTemplate("switch", "expr", "goalState");
            addContents(getPatternMatchCost, switchOnGoal);

            for ( String nonterminal: patternMatches.keySet() )
            {
                ArrayList<JBurgProduction> currentPatterns = patternMatches.get(nonterminal);

                StringTemplate switchCase = getTemplate("switchCase", "label", getUndecoratedNonterminal(nonterminal));
                addContents(switchOnGoal, switchCase);

                // TODO: If there's only one alternative, then just call getCost()
                // to use the cached cost if it's present.
                addAssignment(switchCase, "result", getCostForProduction(currentPatterns.get(0), productions));

                for ( int i = 1; i < currentPatterns.size(); i++ )
                    addAssignment(switchCase, "result", genMinimum("result", getCostForProduction(currentPatterns.get(i), productions)));
            }
            addReturnValue(getPatternMatchCost, "result");
        }

        // getClosureCost(nonterminal)
        ArrayList<String> closureCaches = new ArrayList<String>();
        if ( closureCosts.keySet().size() > 0 )
        {
            StringTemplate getClosureCost = getTemplate(
                "method",
                "decl", codeEmitter.declareMethod(
                    Modifier.PRIVATE,
                    "int",
                    "getClosureCost",
                    genFormals(this.ntType, "goalState"),
                    throwsNothing
                )
            );

            addContents(annotationClass, getClosureCost);
            addVariableDecl(getClosureCost, "int", "result", codeEmitter.genMaxIntValue());

            StringTemplate switchOnGoal = getTemplate("switch", "expr", "goalState");
            addContents(getClosureCost, switchOnGoal);
            
            for ( String nonterminal : closureCosts.keySet() )
            {
                StringTemplate switchCase = getTemplate("switchCase", "label", getUndecoratedNonterminal(nonterminal));
                addContents(switchOnGoal, switchCase);

                Set<String> doneNts = new HashSet<String>();
                List<String> closureComputations = new ArrayList<String>();

                for ( ClosureRecord closure: closureCosts.get(nonterminal) )
                    closureComputations.addAll(getAntecedentCosts(closure, doneNts, closureCosts, patternMatches));

                if ( doneNts.contains(nonterminal) )
                    logger.error(String.format("cyclical closure from %s to %s"));

                // Don't emit pattern match costs for the final level;
                // they're redundant with the computation in getCost().
                closureComputations.add(getClosureCostComputation(nonterminal, closureCosts, patternMatches, false));

                boolean cacheClosure = closureComputations.size() > this.closureCacheThreshold;
                String closureCacheVar = String.format("cachedClosure_%x", System.identityHashCode(closureComputations));

                StringTemplate closureContainer = switchCase;

                if ( cacheClosure )
                {
                    closureCaches.add(closureCacheVar);
                    closureContainer = getTemplate(
                        "ifThen",
                        "condition", getTemplate("equals", "lhs", closureCacheVar, "rhs", UNINITIALIZED)
                    );
                    addContents(switchCase, closureContainer);
                }

                for ( String s: closureComputations )
                    addContents(closureContainer, s);

                if ( cacheClosure )
                {
                    addAssignment(closureContainer, closureCacheVar, String.format("cost_%s", nonterminal));
                    addAssignment(switchCase, "result", closureCacheVar);
                }
                else
                {
                    addAssignment(switchCase, "result", String.format("cost_%s", nonterminal));
                }
            }

            addReturnValue(getClosureCost, "result");
        }

        for ( String closureCacheVar: closureCaches )
        {
            addContents(annotationClass, codeEmitter.genInstanceField(Modifier.PRIVATE, "int", closureCacheVar, UNINITIALIZED));
        }

        // getCost(nonterminal)
        StringTemplate getCost = addSubTemplate(
            annotationClass,
            "method",
            "decl", codeEmitter.declareMethod(
                Modifier.PUBLIC,
                "int",
                "getCost",
                genFormals(this.ntType, "goalState"),
                throwsNothing
            )
        );

        addVariableDecl(getCost, "int", "result", codeEmitter.genMaxIntValue());

        StringTemplate switchOnGoal = getTemplate("switch", "expr", "goalState");
        addContents(getCost, switchOnGoal);
        for ( String nonterminal: productions.keySet() )
        {
            StringTemplate switchCase = addSubTemplate(
                switchOnGoal, 
                "switchCase",
                "label", getUndecoratedNonterminal(nonterminal)
            );

            Object cost;
            StringTemplate callGetPatternMatchCost = getTemplate(
                "callMethod",
                "nameElements", "getPatternMatchCost",
                "params", getNonterminal(nonterminal)
            );
            StringTemplate callGetClosureCost = getTemplate(
                "callMethod",
                "nameElements", "getClosureCost",
                "params", getNonterminal(nonterminal)
            );

            if ( closureCosts.containsKey(nonterminal) && patternMatches.containsKey(nonterminal) )
            {
                cost = genMinimum(callGetPatternMatchCost, callGetClosureCost);
            }
            else if ( patternMatches.containsKey(nonterminal) )
            {
                cost = callGetPatternMatchCost;
            }
            else if ( closureCosts.containsKey(nonterminal) )
            {
                cost = callGetClosureCost;
            }
            else
            {
                // Semantically dead code, compiler appeasement
                throw new IllegalStateException("No closure or pattern cost for " + nonterminal);
            }

            if ( cachedCosts.containsKey(nonterminal) )
            {
                String cachedCostVar = cachedCosts.get(nonterminal);

                // Put the cost into the cache if it's not already done.
                StringTemplate checkCache = addSubTemplate(
                    switchCase,
                    "ifThen",
                    "condition", getTemplate("equals", "lhs", cachedCostVar, "rhs", UNINITIALIZED)
                );

                addAssignment(checkCache, cachedCostVar, cost);
                addAssignment(switchCase, "result", cachedCostVar);
            }
            else
            {
                addAssignment(switchCase, "result", cost);
            }

        }

        addReturnValue(getCost, "result");

        /*
         *  ** Emit getRule()  **
         */
        StringTemplate getRule = addSubTemplate(
            annotationClass,
            "method",
            "decl", codeEmitter.declareMethod(
                Modifier.PUBLIC,
                "int",
                "getRule",
                genFormals(this.ntType, "goalState"),
                throwsNothing
            )
        );

        addVariableDecl(getRule, "int", "rule", NO_FEASIBLE_RULE);
        switchOnGoal = addSubTemplate(
            getRule,
            "switch",
            "expr", "goalState"
        );

        for ( Map.Entry<String, ArrayList<JBurgProduction>> productionsByNonterminal: productions.entrySet() )
        {
            StringTemplate switchCase = addSubTemplate(
                switchOnGoal,
                "switchCase",
                "label", getUndecoratedNonterminal(productionsByNonterminal.getKey())
            );

            ArrayList<JBurgProduction> currentProductions = productionsByNonterminal.getValue();

            /*
             * Try to resolve the optimal production at compiler-compile time.
             */
            JBurgProduction optimalProduction = findOptimalProduction(currentProductions, productions);

            if ( optimalProduction != null )
            {
                addAssignment(switchCase, "rule", Integer.toString(optimalProduction.getReduceAction().getIndex()));
            }
            else
            {
                /*
                 * Emit the analogous compile-time computation.
                 */
                int nProductions = currentProductions.size();

                //  Emit this declaration at its first use.
                boolean currentCostDeclared = false;

                // Cache the best cost so far if necessary.
                Object bestKnownCost;

                if ( nProductions > 1 )
                {
                    addVariableDecl(switchCase, "int", "bestCost", codeEmitter.genMaxIntValue());
                    bestKnownCost = "bestCost";
                }
                else
                {
                    bestKnownCost = codeEmitter.genMaxIntValue();
                }

                for ( int i = 0; i < nProductions; i++ )
                {
                    boolean costIsVariable = true;
                    String currentProductionCost;

                    //  Extract required information from the production:
                    //  what is its cost, and is it constant?
                    JBurgProduction production = currentProductions.get(i);

                    if ( production.computesConstantCost(productions) )
                    {
                        int constantCost = production.getConstantCost(productions);
                        costIsVariable = constantCost == Integer.MAX_VALUE;
                        currentProductionCost = Integer.toString(constantCost);
                    }
                    else
                    {
                        currentProductionCost = getCostForProduction(production, productions);
                    }

                    //  Generate the necessary tests and assignments.

                    //  If the first production's cost is constant
                    //  (and feasible, which has been checked and
                    //  incorporated into costIsVariable), then 
                    //  testing it against bestCost is a tautology.
                    boolean needComparison = (costIsVariable) || i > 0;

                    StringTemplate costContainer = switchCase;

                    if ( needComparison )
                    {
                        //  If the cost uses a computation, put it in a temp.
                        if ( costIsVariable )
                        {
                            if ( currentCostDeclared )
                            {
                                addAssignment(switchCase, "currentCost", currentProductionCost);
                                currentProductionCost = "currentCost";
                            }
                            else if ( nProductions > 1 && costIsVariable )
                            {
                                addVariableDecl(switchCase, "int", "currentCost", currentProductionCost);
                                currentCostDeclared = true;
                                currentProductionCost = "currentCost";
                            }
                            // else fall through using the cost in-line for the comparison.
                        }

                        costContainer = addSubTemplate(
                            costContainer,
                            "ifThen",
                            "condition", codeEmitter.genCmpLess(bestKnownCost.toString(), currentProductionCost)
                        );
                    }

                    //  Track the new best cost if there's another choice to be evaluated.
                    if ( i + 1 < currentProductions.size() )
                        addAssignment(costContainer, bestKnownCost, currentProductionCost);

                    addAssignment(costContainer, "rule", production.getReduceAction().getIndex());
                }
            }
        }

        addReturnValue(getRule, "rule");

        /*
         *  ** Emit getArity()  **
         */
        StringTemplate getArity = addSubTemplate(
            annotationClass, 
            "method",
            "decl", codeEmitter.declareMethod(
                Modifier.PUBLIC,
                "int",
                "getArity",
                noFormalParameters,
                throwsNothing
            )
        );

        if ( fixedArity )
        {
            addReturnValue(getArity, Integer.toString(nominalArity));
        }
        else 
        {
            StringTemplate narySize = getTemplate("narySubtreeSize", "stem", "narySubtrees");

            if ( nominalArity != 0 )
            {
                addReturnValue(getArity, getTemplate("add", "lhs", nominalArity, "rhs", narySize));
            }
            else
            {
                addReturnValue(getArity, narySize);
            }
        }


        /*
         *  **  Emit overrides for getNthChild() and addChild() if necessary  **
         */
        if ( nominalArity > 0 || ! fixedArity )
        {
            //  getNthChild
            StringTemplate getNthChild = addSubTemplate(
                annotationClass,
                "method",
                "decl", codeEmitter.declareMethod(
                    Modifier.PUBLIC,
                    "JBurgAnnotation",
                    "getNthChild",
                    genFormals("int", "index"),
                    throwsNothing
                )
            );

            addVariableDecl(getNthChild, codeEmitter.getAnnotationType(), "result", getTemplate("nullPointer"));
            StringTemplate switchOnIndex = addSubTemplate(getNthChild, "switch", "expr", "index");

            for ( int i = 0; i < nominalArity; i++ )
            {
                StringTemplate switchCase = addSubTemplate(switchOnIndex, "switchCase", "label", new Integer(i));
                addAssignment(switchCase, "result", String.format("subtree%d", i));
            }

            if ( ! fixedArity )
            {
                StringTemplate switchDefault = addSubTemplate(switchOnIndex, "switchDefault");
                if ( nominalArity == 0 )
                    addSubTemplate(switchDefault,
                        "assign",
                        "lvar", "result",
                        "rvalue", getTemplate("narySubtreeGet", "stem", "narySubtrees", "index", "index")
                    );
                else
                    addSubTemplate(switchDefault,
                        "assign",
                        "lvar", "result",
                        "rvalue", getTemplate(
                            "narySubtreeGet",
                            "stem", "narySubtrees",
                            "index", getTemplate("subtract", "lhs", "index", "rhs", new Integer(nominalArity))
                        )
                    );
            }
            else
            {
                StringTemplate switchDefault = addSubTemplate(switchOnIndex, "switchDefaultNoFallthrough");
                // TODO: String literal and concat
                addThrow(switchDefault, "\"Invalid index \" + index");
            }

            addReturnValue(getNthChild, "result");

            //  addChild
            StringTemplate addChild = addSubTemplate(
                annotationClass,
                "method",
                "decl", codeEmitter.declareMethod(
                    Modifier.PUBLIC,
                    "void",
                    "addChild",
                    genFormals("JBurgAnnotation", "child"),
                    throwsNothing
                )
            );

            // We're going to compose a sequence
            // of operations, each extending the
            // previous operation.
            StringTemplate previousOp = addChild;

            if ( nominalArity > 0 )
            {
                for ( int i = 0; i < nominalArity; i++ )
                {
                    Object currentChild = String.format("subtree%d", i);
                    StringTemplate isNull = getTemplate("isNull", "expr", currentChild);
                    if ( i == 0 )
                    {
                        StringTemplate ifThen = getTemplate("ifThen", "condition", isNull);
                        addContents(previousOp, ifThen);
                        previousOp = ifThen;
                    }
                    else
                    {
                        StringTemplate elseif = getTemplate("ifElseArm", "condition", isNull);
                        previousOp.setAttribute("elseArm", elseif);
                        previousOp = elseif;
                    }
                    addAssignment(previousOp, currentChild, "child");
                }

                StringTemplate finalElse = getTemplate("ifElseArm");
                previousOp.setAttribute("elseArm", finalElse);
                previousOp = finalElse;
            }

            if ( ! fixedArity )
                addMethodCall(previousOp, "narySubtrees", "add", "child");
            else
                addThrow(previousOp, "\"too many children\"");
        }

        //  Emit caches for any costs that require a function call;
        //  the BURM's contract says these functions are only called once.
        Set<String> emittedCosts = new HashSet<String>();

        for ( ArrayList<ClosureRecord> closures: closureCosts.values() )
            for ( ClosureRecord closure: closures )
                if ( closure.hasCostFunction() && emittedCosts.add(closure.getCachedCost()) )
                    addCachedCost(annotationClass, closure.getCachedCost(), closure.getCost("m_node")); 

        for ( JBurgRule rule: currentRules )
        {
            if ( rule.needsCostFunction() && rule.isStableCostFunction() )
                addGetCostForRule(annotationClass, rule, factored_variables);

            if ( rule.hasCostFunction() && rule.isStableCostFunction() && emittedCosts.add(rule.getCachedCost()) )
                addCachedCost(annotationClass, rule.getCachedCost(), rule.getCost("m_node")); 
        }

        //  Write out the compressed annotation's class declaration.
        output.println(annotationClass);
    }

    /**
     *  @param productions - a list of productions.
     *  @return true if the list has one member and that member has a constant cost.
     */
    boolean hasConstantCost(ArrayList<JBurgProduction> productions)
    {
        return productions != null &&
            productions.size() == 1 && 
            productions.get(0) instanceof JBurgRule &&
            ((JBurgRule)productions.get(0)).hasConstantCost();
    }

    /**
     *  @return the expression to use for a production's cost;
     *    this may be an arithmetic expression or a call
     *    to an implementation method, depending on
     *    the production's complexity.
     */
    String getCostForProduction(JBurgProduction production, Multimap<String,JBurgProduction> productions)
    {
        if ( production instanceof JBurgRule )
        {
            JBurgRule rule = (JBurgRule) production;
            if ( rule.needsCostFunction() )
            {
                if ( rule.isStableCostFunction() )
                    return this.codeEmitter.genCallMethod(null, getCostingFunctionForRule(rule), "goalState");
                else
                    return rule.getCost("getNode()");
            }
            else
            {
                return getCompositeCostOfRule(rule);
            }
        }
        else
        {
            return ((ClosureRecord)production).getCostComputation(productions);
        }
    }

    /**
     *  Do compiler-compile time dynamic programming to choose
     *  the best alternative from a list of possibilities.
     *  @param currentProductions - the list of productions to choose from.
     *  @param allProductions - the entire set of productions active at the site.
     */
    JBurgProduction findOptimalProduction(List<JBurgProduction> currentProductions, Multimap<String, JBurgProduction> allProductions)
    {
        int bestCost = Integer.MAX_VALUE;
        JBurgProduction result = null;

        for ( JBurgProduction production: currentProductions )
        {
            if ( production.computesConstantCost(allProductions) )
            {
                int cost = production.getConstantCost(allProductions);
                if ( cost < bestCost )
                {
                    result = production;
                    bestCost = cost;
                }
            }
            else
            {
                //  Can't be determined at compiler-compile time.
                result = null;
                break;
            }
        }

        return result;
    }

    /**
     *  Emit a rule's cost function.
     *  @param output - the current output stream.
     *  @param rule - the rule to emit.
     *  @param factored_variables - the list of pattern matchers it's worth factoring.
     */
    void addGetCostForRule(StringTemplate container, JBurgRule rule, Map<JBurgPatternMatcher,String> factored_variables )
    {
        assert(rule.needsCostFunction());

        StringTemplate method = addSubTemplate(
            container,
            "method",
            "decl", codeEmitter.declareMethod(
                Modifier.PRIVATE,
                "int",
                getCostingFunctionForRule(rule),
                genFormals(this.ntType, "goalState"),
                throwsNothing
            )
        );

        for ( String var_decl: factored_variables.values() )
            if ( rule.patternMatcher.usesFactoredVariable(var_decl) )
                addContents(method, var_decl);

        String patternMatch = rule.patternMatcher.generatePatternRecognizer(
                codeEmitter,
                "this",
                JBurgGenerator.this.adapter2
            );

        // The pattern match may be null if it's trival.
        StringTemplate matcher;

        if ( patternMatch != null )
        {
            matcher = addSubTemplate(
                method,
                "ifThen",
                "condition", patternMatch
            );

            StringTemplate failedMatch = getTemplate("ifElseArm");
            addReturnValue(failedMatch, codeEmitter.genMaxIntValue());
            matcher.setAttribute("elseArm", failedMatch);
        }
        else
        {
            matcher = method;
        }

        addReturnValue(matcher, getCompositeCostOfRule(rule));
    }

    /**
     *  @return a unique identifier for the method that computes
     *    a rule's cost (note: this is not the rule's cost function).
     */
    String getCostingFunctionForRule(JBurgRule rule)
    {
        if ( ! ruleNumber.containsKey(rule) )
            ruleNumber.put(rule, String.format("getCostForRule%x", ruleNumber.size()));

        return ruleNumber.get(rule);
    }

    private Map<JBurgRule,String> ruleNumber = new HashMap<JBurgRule, String>();

    /**
     *  Emit the cache field and accessor function for
     *  the cached result of a cost function call.
     */
    void addCachedCost(StringTemplate container, String functionName, String payload)
    {
        //  Strip () characters.
        functionName = functionName.substring(0, functionName.length() -2);

        String varName = String.format("cachedCostFunctionResult_%h", functionName);

        addContents(container, codeEmitter.genInstanceField( Modifier.PRIVATE, "int", varName, UNINITIALIZED));

        StringTemplate method = addSubTemplate(
            container,
            "method",
            "decl", codeEmitter.declareMethod(Modifier.PRIVATE, "int", functionName, noFormalParameters, throwsNothing)
        );

        StringTemplate testInit = addSubTemplate(
            method,
            "ifThen",
            "condition", codeEmitter.genCmpEquality(varName, UNINITIALIZED, true)
        );
        addAssignment(testInit, varName, payload);
        addReturnValue(method, varName);
    }

    /**
     *  Add up all of a rule's cost factors.
     *  @param rule - the rule of interest.
     *  @return an overflow-safe addition of the
     *    rule's cost and the costs of its subtrees.
     */
    String getCompositeCostOfRule(JBurgRule rule)
    {
        String subtreeCost = null;

        for ( JBurgPatternMatcher subgoal: rule.patternMatcher.getParameterizedSubtrees() )
        {
            String subgoalCost = subgoal.generateCost(codeEmitter, "this");

            if ( subgoalCost != null )
                subtreeCost = codeEmitter.genOverflowSafeAdd(subtreeCost, subgoalCost);
        }

        if ( subtreeCost != null )
            return codeEmitter.genOverflowSafeAdd(rule.getCachedCost("getNode()"), subtreeCost);
        else
            return rule.getCachedCost();
    }

    /**
     *  Emit the factored path variables into a map,
     *  and mutate the affected pattern matchers
     *  to refer to the factored variable.
     *  @param commonSubtrees - the map of common subtrees to their pattern matchers.
     */
    Map<JBurgPatternMatcher,String> emitFactoredPathVariables(Multimap<JBurgPatternMatcher, JBurgPatternMatcher> commonSubtrees)
    {
        Map<JBurgPatternMatcher,String> result = new TreeMap<JBurgPatternMatcher, String>();

        int varNum = 0;
        for ( Map.Entry<JBurgPatternMatcher, ArrayList<JBurgPatternMatcher>> factored: commonSubtrees.entrySet() )
        {
            String varName = String.format("factoredPath_%d", varNum++);

            result.put(factored.getKey(), codeEmitter.genLocalVar("JBurgAnnotation", varName, factored.getKey().generateFactoredReference(codeEmitter)));

            for ( JBurgPatternMatcher matcher: factored.getValue() )
                matcher.factoredPath = varName;
        }
    
        return result;
    }

    /**
     *  ClosuresByNonterminal is a convenience class that holds
     *  closure sets grouped by nonterminal; the NT may
     *  be the production or the antecedent, depending on usage.
     */
    class ClosuresByNonterminal extends Multimap<String, ClosureRecord>
    {
        /**
         *  Add a closure record.
         *  @param nt - the nonterminal that indexes this closure.
         *    May be the production or the antecedent.
         */
        public void addClosure(String nt, ClosureRecord closure)
        {
            if ( ! this.getSet(nt).contains(closure) )
                this.getSet(nt).add(closure);
        }
    }

    /**
     *  A ClosureGraph holds the graph of closures that
     *  can be reached from a starting nonterminal.
     */
    class ClosureGraph extends ClosuresByNonterminal
    {
        /**
         *  Sweep the set of nonterminal-to-nonterminal rules and
         *  add any that can be produced by the most recently
         *  added nonterminal.
         *  @param currentNT - the most recently added nonterminal.
         */
        void addClosures(String currentNT)
        {
            addClosures(currentNT, currentNT);
        }

        /**
         *  Sweep the set of nonterminal-to-nonterminal rules and
         *  add any that can be produced by the most recently
         *  added nonterminal.
         *  @param currentNT - the most recently added nonterminal.
         *  @param patternNT - the nonterminal produced by the pattern match.
         */
        void addClosures(String currentNT, String patternNT)
        {
            if ( JBurgGenerator.this.closureSets.containsKey(currentNT) )
            {
                for ( ClosureRecord closure: JBurgGenerator.this.closureSets.get(currentNT) )
                {
                    String newNT = closure.getGoalState();

                    //  Can't replace the pattern with a closure.
                    if ( !newNT.equals(patternNT) )
                    {
                        if ( ! this.containsKey ( newNT ) )
                        {
                            super.addClosure(newNT, closure);
                            addClosures(newNT);
                        }
                        else
                        {
                            //  Add this closure, but its consequent
                            //  closures are already in the set so
                            //  it's not necessary to add them.
                            super.addClosure(newNT, closure);
                        }
                    }
                }
            }
        }
    }

    /**
     *  @return the current Logger implementation;
     *    constructs a Logger that emits messages
     *    to stdout/stderr.
     */
    Logger getLogger()
    {
        if ( null == this.logger )
            this.logger = new Logger(true, true, true);
        return this.logger;
    }

    /**
     *  Convenience method wraps codeEmitter.genAssignment and prints the result.
     */
	void genAssignment(PrintStream output, String lvalue, String rvalue)
	{
		genLine(output, codeEmitter.genAssignment( lvalue, rvalue ) );
	}

    /**
     *  Convenience method wraps codeEmitter.genComment and prints the result.
     */
    void genComment(PrintStream output, String comment)
    {
        genLine(output, codeEmitter.genComment(comment) );
    }

    /**
     *  Convenience method wraps codeEmitter.genLine and prints the result.
     *  @param output - the current output stream
     *  @param contents - the contents to write.  Contents will be trimmed
     *    of whitespace before output.
     */
    void genLine(PrintStream output, String contents)
    {
        String formattedContents = contents.trim();
        output.print(codeEmitter.genLine(formattedContents) );
    }

    /**
     * Convenience method formats and prints a line.
     * @param output the current output stream.
     * @param format a printf-style format string.
     * @param args arguments for printf.
     */
    void printf(PrintStream output, String format, Object ... args)
    {
        genLine(output, String.format(format, args));
    }

    /**
     *  Convenience method wraps codeEmitter.genBeginBlock and prints the result.
     */
    void genBeginBlock(PrintStream output)
    {
        output.print(codeEmitter.genBeginBlock());
    }
    
    /**
     *  Convenience method wraps codeEmitter.genEndBlock and prints the result.
     */
    void genEndBlock(PrintStream output)
    {
        output.print(codeEmitter.genEndBlock());
    }

    /**
     *  Convenience method wraps codeEmitter.genSwitch and prints the result.
     */
    void genSwitch(PrintStream output, String expression)
    {
        genLine(output, codeEmitter.genSwitch(expression) );
        genBeginBlock(output);
    }

    /**
     *  Convenience method wraps codeEmitter.genEndSwitch and prints the result.
     */
    void genEndSwitch(PrintStream output)
    {
        genLine(output, codeEmitter.genEndSwitch() );
    }

    /**
     *  Convenience method wraps codeEmitter.genCase and prints the result.
     */
    void genCase(PrintStream output, String selector )
    {
        output.print(codeEmitter.genCase(selector));
    }

    /**
     *  Convenience method wraps codeEmitter.genEndCase and prints the result.
     */
    void genEndCase(PrintStream output)
    {
        output.print(codeEmitter.genEndCase());
    }

    /**
     *  Convenience method wraps codeEmitter.genDefaultCase and prints the result.
     */
    void genDefaultCase(PrintStream output)
    {
        genLine(output, codeEmitter.genDefaultCase());
        genBeginBlock(output);
    }

    /**
     *  Convenience method wraps codeEmitter.genIf and prints the result.
     */
    void genIf(PrintStream output, String condition )
    {
        genLine(output, codeEmitter.genIf(condition));
    }

    /**
     *  Convenience method wraps codeEmitter.genReturnValue and prints the result.
     */
    void genReturnValue(PrintStream output, String value )
    {
        genLine(output, codeEmitter.genReturnValue(value) + codeEmitter.genEndStmt());
    }

    /**
     *  Convenience method prints an expression as a statement.
     */
    void genExpressionStmt(PrintStream output, String expr)
    {
        genLine(output, expr + codeEmitter.genEndStmt());
    }

    /**
     *  Convenience method wraps codeEmitter.genDeclareMethod and prints the result.
     */
    void genDeclareMethod(PrintStream output, int modifiers, String returnClass, String name, String[][] plist, Class<?>[] exceptions )
    {
        output.print(codeEmitter.declareMethod(modifiers, returnClass, name, plist, exceptions));
    }

    /**
     *  Convenience method declares BURM methods that don't throw.
     */
    void genDeclareMethod(PrintStream output, int modifiers, String returnClass, String name, Object ... plist)
    {
        genDeclareMethod(output, modifiers, returnClass, name, genFormals(plist), throwsNothing);
    }

    /**
     *  Convenience method declares BURM methods that don't throw or have parameters.
     */
    void genDeclareMethod(PrintStream output, int modifiers, String returnClass, String name)
    {
        genDeclareMethod(output, modifiers, returnClass, name, noFormalParameters, throwsNothing);
    }

    /**
     *  Generate formal parameters from a list of type, name pairs.
     *  @param raw_formals - a variadic list of type, name pairs.
     *  @return the raw formals marshalled into a 2-dimensional array.
     */
    String[][] genFormals(Object ... raw_formals)
    {
        assert(raw_formals.length % 2 == 0): "n-ary formal parameters must be in (type, name) pairs";

        String[][] plist = new String[raw_formals.length/2][2];
        for ( int i = 0; i < raw_formals.length - 1; i += 2  )
        {
            plist[i/2][0] = raw_formals[i].toString();
            plist[i/2][1] = raw_formals[i+1].toString();
        }

        return plist;
    }

    /**
     *  Convenience method wraps codeEmitter.genLocalVar and prints the result.
     */
    void genLocalVar(PrintStream output, String varType, String varName, String initializer)
    {
        genLine(output, codeEmitter.genLocalVar(varType, varName, initializer));
        output.println();
    }

    /**
     *  Convenience method wraps codeEmitter.genLocalVar and prints the result.
     */
    void genLocalVar(PrintStream output, String varType, String varName)
    {
        genLocalVar(output, varType, varName, null);
    }

    /**
     *  Convenience method wraps codeEmitter.genThrow and prints the result.
     */
    void genThrow( PrintStream output, String diagnostic )
    {
        genLine(output, codeEmitter.genThrow(diagnostic) + codeEmitter.genEndStmt() );
    }

    /**
     *  Convenience method wraps codeEmitter.genInstanceField and prints the result.
     */
    void genInstanceField(PrintStream output, int modifiers, String type, String name, String initializer)
    { 
        genLine(output, codeEmitter.genInstanceField(modifiers, type, name, initializer));
    }

    /**
     *  Convenience method wraps codeEmitter.genGetGoalState.
     */
    String getNonterminal(String raw_nt)
    {
        return codeEmitter.genGetGoalState(raw_nt);
    }

    /**
     *  Convenience method wraps codeEmitter.genGetUndecoratedGoalState.
     */
    String getUndecoratedNonterminal(String raw_nt)
    {
        return codeEmitter.genGetUndecoratedGoalState(raw_nt);
    }

    /**
     *  @return the input string with its outer
     *    layer of {} brackets stripped off.
     */
    public static String stripBrackets(String src)
    {
        int startchar = src.indexOf('{') + 1;
        int lentoend = src.lastIndexOf('}');
        return src.substring( startchar, lentoend-startchar );
    
    }

    /**
     *  Generate the correct calling sequence to get the nth child
     *  of an input i-node.
     *  @param indexTerm - the n in nth child.
     */
    String genGetNthChild(String indexTerm)
    {
		return this.adapter2 != null ?
		    this.codeEmitter.genCallMethod("this", "getNthChild", JBurgGenerator.initialParamName, indexTerm):
	        this.iNodeAdapter.genGetNthChild( JBurgGenerator.initialParamName, indexTerm, codeEmitter );
    }
		
    /**
     *  Generate the calling sequence for the label() method.
     */
    String genCallLabel(String indexTerm)
    {
        return this.codeEmitter.genCallMethod(
            "this",
            "label",
            codeEmitter.genCast( 
                iNodeClass, 
                genGetNthChild(indexTerm)
            )
        );
    }

    /**
     * Generate code to check for a minumum of two values.
     * @param lhs the left-hand expression.
     * @param rhs the right-hand expression.
     * @return platform-specific code to get the minimum of these values.
     */
    String genMinimum(Object lhs, Object rhs)
    {
        if ( lhs.equals(rhs) && this.volatileCostFunctions.size() == 0 )
            return lhs.toString();
        else
            return getTemplate("minimumOf", "e1", lhs, "e2", rhs).toString();
    }

    /**
     *  @return true if any rules in the list have n-ary operands.
     */
    private boolean hasNaryness(Collection<JBurgRule> rules)
    {
        boolean result = false;

        for ( JBurgRule rule: rules )
            result |= rule.patternMatcher.hasNaryness();

        return result;
    }

    /**
     *  @return the minumum arity of a list of rules.
     */
    private int getMinumumArity(Collection<JBurgRule> rules)
    {
        int result = Integer.MAX_VALUE;

        for ( JBurgRule rule: rules )
            result = Math.min(result, rule.patternMatcher.getNominalArity());

        return result;
    }

    /**
     *  @return the name of a JBurgAnnotation specialized subclass.
     */
    String getSpecializedClassName(ArrayList<JBurgRule> rules )
    {
        if ( hasNaryness(rules) )
            return String.format("JBurgAnnotation_%s_%d_n", rules.get(0).getOperator(), getMinumumArity(rules));
        else
            return String.format("JBurgAnnotation_%s_%d", rules.get(0).getOperator(), getMinumumArity(rules));
    }

    /**
     *  RulesByOperatorAndArity is a multidimensional associative store that
     *  sorts JBurgRules into equivalence classes, where the equivalaence
     *  relation is "these rules can share an annotation object."
     */
    class RulesByOperatorAndArity
    {
        /**
         *  Unsorted rules.
         */
        Multimap<String, JBurgRule> unsortedRules = new Multimap<String, JBurgRule>();

        Map<String, Iterable<ArrayList<JBurgRule>>> sortedRules = null;

        public void addAll(List<JBurgRule> rules)
        {
            assert sortedRules == null;

            String operator = rules.get(0).getOperator();
            this.unsortedRules.addAllToSet(operator, rules);
        }

        public void addRule(JBurgRule rule)
        {
            assert sortedRules == null;
            this.unsortedRules.addToSet(rule.getOperator(), rule);
        }

        public Set<String> getOperators()
        {
            return unsortedRules.keySet();
        }

        public Iterable<ArrayList<JBurgRule>> getRulesFor(String operator)
        {
            if ( this.sortedRules == null )
            {
                this.sortedRules = new TreeMap<String, Iterable<ArrayList<JBurgRule>>>();

                for ( String op: getOperators() )
                    this.sortedRules.put(op, sortRules(this.unsortedRules.get(op)));
            }

            return this.sortedRules.get(operator);
        }

        private Iterable<ArrayList<JBurgRule>> sortRules( ArrayList<JBurgRule> unsorted_rules)
        {
            //  Find the minumum arity of all n-ary patterns;
            //  any rule with arity >= this limit gets sorted
            //  into the "variable-arity" bucket.
            int arity_requires_variable = Integer.MAX_VALUE;

            for ( JBurgRule rule: unsorted_rules )
                if ( rule.patternMatcher.hasNaryness() )
                    arity_requires_variable = Math.min(arity_requires_variable, rule.patternMatcher.getNominalArity());

            Multimap<Integer, JBurgRule> rules = new Multimap<Integer, JBurgRule>(); 

            for ( JBurgRule rule: unsorted_rules )
            {
                //  All n-ary patterns and any fixed-arity patterns that
                //  overlap with n-ary patterns go into a common annotation;
                //  fixed-arity patterns of arity less than the smallest
                //  n-ary arity can't be confused with an n-ary pattern
                //  and can go in their own annotation.
                Integer key = 
                    rule.patternMatcher.getNominalArity() < arity_requires_variable?
                    rule.patternMatcher.getNominalArity():
                    arity_requires_variable;

                rules.addToSet(key,rule);
            }

            return rules.values();
        }
    }

    /**
     * Translate a JBurg platform-neutral expression
     * into platform-specific code.
     */
    private String emitExpression(AST expr)
    {
        ExpressionParser parser = new ExpressionParser();
        parser.setCodeEmitter(this.codeEmitter);
        try
        {
            parser.burm(expr);
        }
        catch ( Exception burmProblem )
        {
            logger.exception(String.format("translating %s", expr), burmProblem);
        }

        return parser.getResult().toString();
    }

    /**
     * Get the cost of a closure as a "term" of a closure cost function.
     * @param nonterminal the goal nonterminal.
     * @param closureCosts the current annotation's graph of closure costs.
     * @param patternMatches the current annotation's pattern matches.
     * @param emitPatternMatch if set, emit a pattern match if the nonterminal
     * has any pattern matches; true when emitting antecedent terms, false
     * when emitting the final term, when the computation would be redundant
     * with the computation in getCost().
     * @return the closure's cost, expressed as an initialization of a cost term.
     */
    String getClosureCostComputation(String nonterminal, ClosureGraph closureCosts, Multimap<String, JBurgProduction> patternMatches, boolean emitPatternMatch)
    {
        List<String> result = new ArrayList<String>();

        String minCost;
        String patternMatchCost = this.codeEmitter.genCallMethod("this", "getPatternMatchCost", getNonterminal(nonterminal));

        //  Determine the minimum cost.
        if ( closureCosts.containsKey(nonterminal) )
        {
            ArrayList<ClosureRecord> possibleClosures = closureCosts.getSet(nonterminal);

            minCost = getClosureCost(possibleClosures.get(0), closureCosts, patternMatches);

            for ( int i = 1; i < possibleClosures.size(); i++ )
                minCost = genMinimum( minCost, getClosureCost(possibleClosures.get(i), closureCosts, patternMatches) );

            if ( patternMatches.containsKey(nonterminal) && emitPatternMatch )
                minCost = genMinimum(minCost, patternMatchCost);
        }
        else
        {
            minCost = patternMatchCost;
        }

        return getTemplate(
            "declareLocalVariable",
            "type", "int",
            "name", String.format("cost_%s", nonterminal), 
            "initializer", minCost
        ).toString();
    }

    /**
     * Get the cost of a closure with its antecedent "term."
     * @param closure the closure of interest.
     * @param closureCosts the current annotation's graph of closure costs.
     * @param patternMatches the current annotation's pattern matches.
     * @return the closure's cost as an expression usable within the annotation class' instance methods.
     */
    String getClosureCost(ClosureRecord closure, ClosureGraph closureCosts, Multimap<String, JBurgProduction> patternMatches)
    {
        return closure.getCostComputation("getNode()", "cost_" + getNonTrivialAntecedent(closure.getAntecedentState(), closureCosts, patternMatches));
    }

    /**
     * Get all the antecedent costs for a closure.
     * @param closure the closure of interest.
     * @param doneNts antecedents already recorded; mutated to record
     * the current closure, if it hasn't already been emitted.
     * @param closureCosts the current annotation's graph of closure costs.
     * @param patternMatches the current annotation's pattern matches.
     * @return a list of all the antecedent "terms" in the closure cost.
     */
    List<String> getAntecedentCosts(ClosureRecord closure, Set<String> doneNts, ClosureGraph closureCosts, Multimap<String, JBurgProduction> patternMatches)
    {
        List<String> result = new ArrayList<String>();

        // Look past any trivial antecedents; if foo = bar is the only antecedent
        // reduction, and it has zero cost, and bar itself has an antecedent, then
        // we're interested in the value of bar.
        // TODO: Also elide the corresponding reduction.
        String antecedent = getNonTrivialAntecedent(closure.getAntecedentState(), closureCosts, patternMatches); 

        if ( doneNts.add(antecedent) )
        {
            if ( closureCosts.containsKey(antecedent) )
                for ( ClosureRecord antecedentClosure: closureCosts.get(antecedent) )
                    result.addAll(getAntecedentCosts(antecedentClosure, doneNts, closureCosts, patternMatches));

            result.add(getClosureCostComputation(antecedent, closureCosts, patternMatches, true));
        }

        return result;
    }

    /**
     * Look through any trivial antecedent states; a trivial
     * antecedent has a single, zero-cost closure and no pattern match.
     * @param antecedent the nonterminal state of interest.
     * @param closureCosts the current annotation's graph of closure costs.
     * @param patternMatches the current annotation's pattern matches.
     * @return the first non-trivial antecedent found.
     */
    String getNonTrivialAntecedent(String antecedent, ClosureGraph closureGraph, Multimap<String, JBurgProduction> patternMatches)
    {
        if (closureGraph.containsKey(antecedent) && !patternMatches.containsKey(antecedent))
        {
            ArrayList<ClosureRecord> closures = closureGraph.get(antecedent);
            if ( closures.size() == 1 && closures.get(0).costIsZero() )
            {
                // One closure with zero cost: it's trivial.
                return getNonTrivialAntecedent(closures.get(0).getAntecedentState(), closureGraph, patternMatches);
            }
        }

        // This is a non-trival closure.
        return antecedent;
    }

    /**
     * Select a template by name, with optional attributes.
     * @param name the template's name.
     * @param attrValue pairs of attrName, attrValue.
     */
    private StringTemplate getTemplate(String name, Object ... attrValue)
    {
        return this.codeEmitter.getTemplate(name, attrValue);
    }

    private StringTemplate addSubTemplate(StringTemplate container, String name, Object ... attrValue)
    {
        StringTemplate contents = getTemplate(name, attrValue);
        addContents(container, contents);
        return contents;
    }

    /**
     * Add a variable declaration to a container's template.
     * @param container the container's template.
     * @param type the variable's type.
     * @param name the variable's name.
     * @param initialValue the variable's initial value, or null.
     */
    private void addVariableDecl(StringTemplate container, Object type, Object name, Object initialValue)
    {
        addFieldOrVariableDecl(container, "declareLocalVariable", type, name, initialValue);
    }

    private void addFieldOrVariableDecl(StringTemplate container, String template,  Object type, Object name, Object initialValue)
    {
        addContents(
            container,
            getTemplate(
                template,
                "type", type,
                "name", name,
                "initializer", initialValue
            )
        );
    }

    /**
     * Add an assignment statement to a container template.
     * @param method the method's template.
     * @param lvalue the receiving lvalue.
     * @param rvalue the value to assign.
     */
    private void addAssignment(StringTemplate method, Object lvalue, Object rvalue)
    {
        addContents(method, getTemplate( "assign", "lvar", lvalue, "rvalue", rvalue));
    }

    /**
     * Add contents to a container template.
     * @param container  the container.
     * @param contents the contents.
     */
    private void addContents(StringTemplate container, Object contents)
    {
        container.setAttribute("contents", contents);
    }

    /**
     * Add a return statement to a container template.
     * @param container the containing template.
     * @param value the value to return.
     */
    private void addReturnValue(StringTemplate container, Object value)
    {
        addContents(container, getTemplate( "returnStmt", "value", value));
    }

    /**
     * Add an expression statement to a container template.
     * @param container the containing template.
     * @param expr the expression to add as a statement-level construct.
     */
    private void addExprStatement(StringTemplate container, Object expr)
    {
        addContents(container, getTemplate("statement", "expr", expr));
    }

    /**
     * Add a method call to a container template.
     * @param container the containing template.
     * @param stem the stem of the method reference.  May be null.
     * @param methodName the name of the method to call.
     * @param params actual parameter expressions.
     */
    private void addMethodCall(StringTemplate container, Object stem, Object methodName, Object ... params)
    {
        StringTemplate methodCall = 
            stem != null?
                getTemplate(
                    "callMethod",
                    "nameElements", stem,
                    "nameElements", methodName,
                    "params", params
                ):
                getTemplate(
                    "callMethod",
                    "nameElements", methodName,
                    "params", params
                )
        ;

        addExprStatement(container, methodCall);
    }

    /**
     * Add a throw statement to a container template.
     * @param container the containing template.
     * @param diagnostic an expression that generates a diagnostic.
     */
    private void addThrow(StringTemplate container, Object diagnostic)
    {
        addExprStatement(container, getTemplate("throwDiagnostic", "diagnostic", diagnostic));
    }
}
